
__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__   = "MIT"


import os as _os


# -----------------------------------------------------------------------------
# common states
NEW      = 'NEW'
DONE     = 'DONE'
FAILED   = 'FAILED'
CANCELED = 'CANCELED'

# shortcut
INITIAL  = [NEW]
FINAL    = [DONE, FAILED, CANCELED]


# ------------------------------------------------------------------------------
#
# pilot states
#
PMGR_LAUNCHING_PENDING = 'PMGR_LAUNCHING_PENDING'
PMGR_LAUNCHING         = 'PMGR_LAUNCHING'
PMGR_ACTIVE_PENDING    = 'PMGR_ACTIVE_PENDING'
PMGR_ACTIVE            = 'PMGR_ACTIVE'

# assign numeric values to states to support state ordering operations
# ONLY final state get the same values.
_pilot_state_values = {
        None                   : -1,
        NEW                    :  0,
        PMGR_LAUNCHING_PENDING :  1,
        PMGR_LAUNCHING         :  2,
        PMGR_ACTIVE_PENDING    :  3,
        PMGR_ACTIVE            :  4,
        DONE                   :  5,
        FAILED                 :  5,
        CANCELED               :  5}

_pilot_state_inv   = {_v: _k for _k, _v in _pilot_state_values.items()}
_pilot_state_inv_full = dict()

for _st,_v in _pilot_state_values.items():
    if _v in _pilot_state_inv_full:
        if not isinstance(_pilot_state_inv_full[_v], list):
            _pilot_state_inv_full[_v] = [_pilot_state_inv_full[_v]]
        _pilot_state_inv_full[_v].append(_st)
    else:
        _pilot_state_inv_full[_v] = _st


# ------------------------------------------------------------------------------
#
def _pilot_state_value(s):
    return _pilot_state_values[s]


# ------------------------------------------------------------------------------
#
def _pilot_state_progress(pid, current, target):
    """See documentation of 'task_state_progress' below."""

    # first handle final state corrections
    if current == CANCELED:
        if target in [DONE, FAILED, CANCELED]:
            return [target, []]

    # allow to transition from FAILED to DONE (done gets picked up from proxy,
    # sometimes after pilot watcher detects demise)
    if current == FAILED:
        if target in FINAL:
            return [target, []]

    if current in FINAL and target != current:
        if target in FINAL:
            raise ValueError('invalid transition for %s: %s -> %s'
                            % (pid, current, target))

    cur = _pilot_state_values[current]
    tgt = _pilot_state_values[target]

    if cur >= tgt:
        # nothing to do, a similar or better progression happened earlier
        return [current, []]

    # dig out all intermediate states, skip current
    passed = list()
    for i in range(cur + 1,tgt):
        passed.append(_pilot_state_inv[i])

    # append target state to trigger notification of transition
    passed.append(target)

    return target, passed


# ------------------------------------------------------------------------------
#
def _pilot_state_collapse(states):
    """Select the relevant state.

    This method takes a list of pilot states and selects the one with the
    highest state value.
    """

    # we first check the final states, as we want to express a preference there.
    # Then we start comparing actual values.
    if DONE     in states: return DONE
    if FAILED   in states: return FAILED
    if CANCELED in states: return CANCELED

    ret = None
    for state in states:
        if _pilot_state_values[state] > _pilot_state_values[ret]:
            ret = state

    return ret


# ------------------------------------------------------------------------------
#
# task states
#
TMGR_SCHEDULING_PENDING      = 'TMGR_SCHEDULING_PENDING'
TMGR_SCHEDULING              = 'TMGR_SCHEDULING'
TMGR_STAGING_INPUT_PENDING   = 'TMGR_STAGING_INPUT_PENDING'
TMGR_STAGING_INPUT           = 'TMGR_STAGING_INPUT'
AGENT_STAGING_INPUT_PENDING  = 'AGENT_STAGING_INPUT_PENDING'
AGENT_STAGING_INPUT          = 'AGENT_STAGING_INPUT'
AGENT_SCHEDULING_PENDING     = 'AGENT_SCHEDULING_PENDING'
AGENT_SCHEDULING             = 'AGENT_SCHEDULING'
AGENT_EXECUTING_PENDING      = 'AGENT_EXECUTING_PENDING'
AGENT_EXECUTING              = 'AGENT_EXECUTING'
AGENT_STAGING_OUTPUT_PENDING = 'AGENT_STAGING_OUTPUT_PENDING'
AGENT_STAGING_OUTPUT         = 'AGENT_STAGING_OUTPUT'
TMGR_STAGING_OUTPUT_PENDING  = 'TMGR_STAGING_OUTPUT_PENDING'
TMGR_STAGING_OUTPUT          = 'TMGR_STAGING_OUTPUT'

# assign numeric values to states to support state ordering operations
# ONLY final state get the same values.
_task_state_values = {
        None                         : -1,
        NEW                          :  0,
        TMGR_SCHEDULING_PENDING      :  1,
        TMGR_SCHEDULING              :  2,
        TMGR_STAGING_INPUT_PENDING   :  3,
        TMGR_STAGING_INPUT           :  4,
        AGENT_STAGING_INPUT_PENDING  :  5,
        AGENT_STAGING_INPUT          :  6,
        AGENT_SCHEDULING_PENDING     :  7,
        AGENT_SCHEDULING             :  8,
        AGENT_EXECUTING_PENDING      :  9,
        AGENT_EXECUTING              : 10,
        AGENT_STAGING_OUTPUT_PENDING : 11,
        AGENT_STAGING_OUTPUT         : 12,
        TMGR_STAGING_OUTPUT_PENDING  : 13,
        TMGR_STAGING_OUTPUT          : 14,
        DONE                         : 15,
        FAILED                       : 15,
        CANCELED                     : 15}
_task_state_inv = {_v: _k for _k, _v in _task_state_values.items()}
_task_state_inv_full = dict()
for _st,_v in _task_state_values.items():
    if _v in _task_state_inv_full:
        if not isinstance(_task_state_inv_full[_v], list):
            _task_state_inv_full[_v] = [_task_state_inv_full[_v]]
        _task_state_inv_full[_v].append(_st)
    else:
        _task_state_inv_full[_v] = _st


def _task_state_value(s):
    return _task_state_values[s]


def _task_state_progress(uid, current, target):
    """Normalize state progression.

    This method will ensure a task state progression in sync with the state
    model defined above.  It will return a tuple: [new_state, passed_states]
    where 'new_state' is either 'target' or 'current', depending which comes
    later in the state model, and 'passed_states' is a sorted list of all states
    between 'current' (excluded) and 'new_state' (included).  If a progression
    is not possible because the 'target' state value is equal or smaller that
    the 'current' state, then 'current' is returned, and the list of passed
    states remains empty.  This way, 'passed_states' can be used to invoke
    callbacks for all state transition, where each transition is announced
    exactly once.

    Note that the call will not allow to transition between states of equal
    values, which in particular applies to final states -- those are truly
    final.  Requesting such transitions will result in silent discard of the
    invalid target state.

    On contradicting final states, DONE and FAILED are preferred over CANCELED,
    but no other transitions are allowed

    task_state_progress(TMGR_SCHEDULING, NEW)
    --> [TMGR_SCHEDULING, []]

    task_state_progress(NEW, NEW)
    --> [NEW, []]

    task_state_progress(NEW, TMGR_SCHEDULING_PENDING)
    --> [TMGR_SCHEDULING_PENDING, [TMGR_SCHEDULING_PENDING]]

    task_state_progress(NEW, TMGR_SCHEDULING)
    --> [TMGR_SCHEDULING, [TMGR_SCHEDULING_PENDING, TMGR_SCHEDULING]]

    task_state_progress(DONE, FAILED)
    --> [DONE, []]

    """

    # first handle final state corrections
    if current == CANCELED:
        if target in [DONE, FAILED, CANCELED]:
            return [target, []]

    if current in FINAL:
        if target in FINAL:
            raise ValueError('invalid transition for %s: %s -> %s'
                             % (uid, current, target))

    cur = _task_state_values[current]
    tgt = _task_state_values[target]

    if cur >= tgt:
        # nothing to do, a similar or better progression happened earlier
        return [current, []]

    # dig out all intermediate states, skip current
    passed = list()
    for i in range(cur + 1,tgt):
        passed.append(_task_state_inv[i])

    # append target state to trigger notification of transition
    passed.append(target)

    return target, passed


# ------------------------------------------------------------------------------
#
def _task_state_collapse(states):
    """Select the highest state value from a list of task states."""

    # we first check the final states, as we want to express a preference there.
    # Then we start comparing actual values.
    if DONE     in states: return DONE
    if FAILED   in states: return FAILED
    if CANCELED in states: return CANCELED
    ret = None
    for state in states:
        if _task_state_values[state] > _task_state_values[ret]:
            ret = state
    return ret


# -----------------------------------------------------------------------------
# backward compatibility
#
# pilot states
if 'RP_ENABLE_OLD_DEFINES' in _os.environ:
    CANCELING              = CANCELED
    PENDING                = PMGR_LAUNCHING_PENDING
    PENDING_LAUNCH         = PMGR_LAUNCHING_PENDING
    LAUNCHING              = PMGR_LAUNCHING
    PENDING_ACTIVE         = PMGR_ACTIVE_PENDING
    ACTIVE_PENDING         = PMGR_ACTIVE_PENDING
    ACTIVE                 = PMGR_ACTIVE

    # task states
    UNSCHEDULED            = TMGR_SCHEDULING_PENDING
    SCHEDULING             = TMGR_SCHEDULING
    PENDING_INPUT_STAGING  = TMGR_STAGING_INPUT_PENDING
    STAGING_INPUT          = TMGR_STAGING_INPUT
    ALLOCATING_PENDING     = AGENT_SCHEDULING_PENDING
    ALLOCATING             = AGENT_SCHEDULING
    PENDING_EXECUTION      = AGENT_EXECUTING_PENDING
    EXECUTING_PENDING      = AGENT_EXECUTING_PENDING
    EXECUTING              = AGENT_EXECUTING
    PENDING_OUTPUT_STAGING = TMGR_STAGING_OUTPUT_PENDING
    STAGING_OUTPUT         = TMGR_STAGING_OUTPUT

_legacy_states = {
    'New'                        : NEW,
    'AllocatingPending'          : TMGR_SCHEDULING_PENDING,
    'Allocating'                 : TMGR_SCHEDULING,
    'PendingInputStaging'        : TMGR_STAGING_INPUT_PENDING,
    'StagingInput'               : TMGR_STAGING_INPUT,
    'AgentStagingInputPending'   : AGENT_STAGING_INPUT_PENDING,
    'AgentStagingInput'          : AGENT_STAGING_INPUT,
    'ExecutingPending'           : AGENT_EXECUTING_PENDING,
    'Executing'                  : AGENT_EXECUTING,
    'AgentStagingOutputPending'  : AGENT_STAGING_OUTPUT_PENDING,
    'AgentStagingOutput'         : AGENT_STAGING_OUTPUT,
    'PendingOutputStaging'       : TMGR_STAGING_OUTPUT_PENDING,
    'StagingOutput'              : TMGR_STAGING_OUTPUT,
    'Done'                       : DONE,
    'Canceled'                   : CANCELED,
    'CANCELED'                   : CANCELED,
    'Failed'                     : FAILED,
    'FAILED'                     : FAILED
}


# backward compatibility

UMGR_SCHEDULING_PENDING      = TMGR_SCHEDULING_PENDING
UMGR_SCHEDULING              = TMGR_SCHEDULING
UMGR_STAGING_INPUT_PENDING   = TMGR_STAGING_INPUT_PENDING
UMGR_STAGING_INPUT           = TMGR_STAGING_INPUT
UMGR_STAGING_OUTPUT_PENDING  = TMGR_STAGING_OUTPUT_PENDING
UMGR_STAGING_OUTPUT          = TMGR_STAGING_OUTPUT


# -----------------------------------------------------------------------------




import os
import sys
import time
import queue

import threading         as mt
import multiprocessing   as mp

import radical.utils     as ru

from .worker  import Worker


# ------------------------------------------------------------------------------
#
class DefaultWorker(Worker):

    # --------------------------------------------------------------------------
    #
    def __init__(self, raptor_id : str):

        # generate a MPI rank dependent UID for each worker process
        # FIXME: this should be delegated to ru.generate_id
        # FIXME: rank determination should be moved to RU
        rank = None

        if rank is None: rank = os.environ.get('PMIX_RANK')
        if rank is None: rank = os.environ.get('PMI_RANK')
        if rank is None: rank = os.environ.get('OMPI_COMM_WORLD_RANK')

        if rank is None: rank = 0
        else           : rank = int(rank)

        # only rank 0 (the manager) registers with the master.
        if rank == 0: manager = True
        else        : manager = False

        self._res_evt = mp.Event()          # set on free resources
        self._my_term = mt.Event()          # for start/stop/join

        super().__init__(manager=manager, rank=rank, raptor_id=raptor_id)

        # connect to the master queues
        self._res_put = ru.zmq.Putter('result',  self._res_addr_put)
        self._req_get = ru.zmq.Getter('request', self._req_addr_get,
                                                 cb=self._request_cb)

        # the master should have stored our own task description in the registry
        self._descr = self._reg['raptor.%s.cfg' % self._uid] or {}

        # keep worker ID and rank
        self._n_cores = int(self._descr.get('cores_per_rank') or
                            os.getenv('RP_CORES_PER_RANK', '1'))
        self._n_gpus  = int(self._descr.get('gpus_per_rank') or
                            os.getenv('RP_GPUS_PER_RANK', '0'))

        # We need to make sure to run only up to `gpn` tasks using a gpu
        # within that pool, so need a separate counter for that.
        self._rlock     = mt.Lock()
        self._resources = {'cores' : [0] * self._n_cores,
                           'gpus'  : [0] * self._n_gpus}

      # self._log.debug('cores %s', str(self._resources['cores']))
      # self._log.debug('gpus  %s', str(self._resources['gpus']))

        # resources are initially all free
        self._res_evt.set()

        self._pool  = dict()     # map task uid to process instance
        self._plock = mt.Lock()  # lock _pool

        # We also create a queue for communicating results back, and a thread to
        # watch that queue
        self._result_queue  = mp.Queue()
        self._result_thread = mt.Thread(target=self._result_watcher)
        self._result_thread.daemon = True
        self._result_thread.start()


    # --------------------------------------------------------------------------
    #
    def start(self):

        pass


    # --------------------------------------------------------------------------
    #
    def stop(self):

        self._my_term.set()


    # --------------------------------------------------------------------------
    #
    def join(self):

        # FIXME
        while True:
            if self._my_term.wait(1):
                break


    # --------------------------------------------------------------------------
    #
    def _alloc(self, task):
        '''
        allocate task resources
        '''

        uid = task['uid']
        self._prof.prof('schedule_try', uid=uid)

        with self._rlock:

            cores = task.get('cores', 1)
            gpus  = task.get('gpus' , 0)

            assert cores >= 1
            assert cores <= self._n_cores
            assert gpus  <= self._n_gpus

            if cores > self._resources['cores'].count(0): return False
            if gpus  > self._resources['gpus' ].count(0): return False

            alloc_cores = list()
            alloc_gpus  = list()

            if cores:
                for n in range(self._n_cores):
                    if not self._resources['cores'][n]:
                        self._resources['cores'][n] = 1
                        alloc_cores.append(n)
                        if len(alloc_cores) == cores:
                            break

            if gpus:
                for n in range(self._n_gpus):
                    if not self._resources['gpus'][n]:
                        self._resources['gpus'][n] = 1
                        alloc_gpus.append(n)
                        if len(alloc_gpus) == gpus:
                            break

            task['slots'] = {'cores': alloc_cores,
                             'gpus' : alloc_gpus}

        self._prof.prof('schedule_ok', uid=uid)

        return True


    # --------------------------------------------------------------------------
    #
    def _dealloc(self, task):
        '''
        deallocate task resources
        '''

        self._prof.prof('unschedule_start', uid=task['uid'])

        with self._rlock:

            resources = task['slots']

            for n in resources['cores']:
                assert self._resources['cores'][n]
                self._resources['cores'][n] = 0

            for n in resources['gpus']:
                assert self._resources['gpus'][n]
                self._resources['gpus'][n] = 0

            # signal available resources
            self._res_evt.set()
            self._prof.prof('unschedule_stop', uid=task['uid'])

        return True


    # --------------------------------------------------------------------------
    #
    def _request_cb(self, tasks):
        '''
        grep call type from tasks, check if methods are registered, and
        invoke them.
        '''

        for task in ru.as_list(tasks):

            task['worker'] = self._uid

            try:

                # ok, we have work to do.  Check the requirements to see how
                # many cpus and gpus we need to mark as busy
                while not self._alloc(task):

                    # no resource - wait for new resources
                    #
                    # NOTE: this will block smaller tasks from being executed
                    #       right now.  alloc_task is not a proper scheduler,
                    #       after all.
                  # while not self._res_evt.wait(timeout=1.0):
                  #     self._log.debug('req_alloc_wait %s', task['uid'])
                  # #   FIXME: `clear` should be locked
                  # self._res_evt.clear()

                    time.sleep(0.01)

                self._prof.prof('req_start', uid=task['uid'], msg=self._uid)

                # we got an allocation for this task, and can run it, so apply
                # to the process pool.  The callback (`self._result_cb`) will
                # pick the task up on completion and free resources.
                #
                # NOTE: we don't use mp.Pool - see __init__ for details

                env = self._task_env
                env['RP_TASK_ID'] = task['uid']

              # ret = self._pool.apply_async(func=self._dispatch, args=[task],
              #                              callback=self._result_cb,
              #                              error_callback=self._error_cb)
                proc = mp.Process(target=self._dispatch, args=(task, env))
              # proc.daemon = True

                with self._plock:

                    # we need to include `proc.start()` in the lock, as
                    # otherwise we may end up getting the `self._result_cb`
                    # before the pid could be registered in `self._pool`.
                    proc.start()
                    self._pool[proc.pid] = proc
                self._log.debug('applied: %s: %s: %s',
                                task['uid'], proc.pid, self._pool.keys())


            except Exception as e:

                self._log.exception('request failed')

                # free resources again for failed task
                self._dealloc(task)

                task['exception']        = repr(e)
                task['exception_detail'] = '\n'.join(ru.get_exception_trace())

                self._res_put.put(task)


    # --------------------------------------------------------------------------
    #
    def _dispatch(self, task, env):

        # this method is running in a process of the process pool, and will now
        # apply the task to the respective execution mode.
        #
        # NOTE: application of pre_exec directives may got here

        task['pid'] = os.getpid()

        # apply task env settings
        for k,v in env.items():
            os.environ[k] = v

        for k,v in task.get('environment', {}).items():
            os.environ[k] = v

        # ----------------------------------------------------------------------
        def _worker_proc(res_lock):
            # FIXME: do we still need this thread?

            import setproctitle
            setproctitle.setproctitle('rp.dispatch.%s' % task['uid'])

            # make CUDA happy
            # FIXME: assume physical device numbering for now
            if task['slots']['gpus']:
                os.environ['CUDA_VISIBLE_DEVICES'] = \
                             ','.join(str(i) for i in task['slots']['gpus'])

            out = None
            err = None
            ret = 1
            val = None
            exc = [None, None]
            try:

                sbox = task['task_sandbox_path']
                ru.rec_makedir(sbox)
                os.chdir(sbox)
                dispatcher = self.get_dispatcher(task['description']['mode'])
                out, err, ret, val, exc = dispatcher(task)

            except Exception as e:
                exc = [repr(e), '\n'.join(ru.get_exception_trace())]

            finally:
                os.chdir(self._sbox)

            res = [task, out, err, ret, val, exc]

            with res_lock:
                self._result_queue.put(res)
        # ----------------------------------------------------------------------


        ret = None
        try:
            tout = task['description']['timeout'] or None

          # self._log.debug('dispatch: %s: %d (%.1f)',
          #                 task['uid'], task['pid'], tout)

            res_lock = mp.Lock()
            worker_proc = mp.Process(target=_worker_proc, args=(res_lock,))
            worker_proc.daemon = True
            worker_proc.start()
            worker_proc.join(timeout=tout)

            with res_lock:
                if worker_proc.is_alive():
                    worker_proc.terminate()
                    worker_proc.join()
                    out = None
                    err = 'timeout (>%s)' % tout
                    ret = 1
                    val = None
                    exc = ['TimeoutError("task timed out")', None]
                    res = [task, str(out), str(err), int(ret), val, exc]
                    self._log.debug('put 2 result: task %s', task['uid'])
                    self._result_queue.put(res)
                    self._log.debug('worker_proc killed: %s', task['uid'])

        except Exception as e:

            self._log.exception('dispatch failed')
            out = None
            err = 'dispatch failed: %s' % e
            ret = 1
            val = None
            exc = [repr(e), '\n'.join(ru.get_exception_trace())]
            res = [task, str(out), str(err), int(ret), val, exc]
            self._log.debug('put 3 result: task %s', task['uid'])

            self._result_queue.put(res)

        finally:
            # if we kill the process too quickly, the result put above
            # will not make it out, thus make sure the queue is empty
            # first.
            ret = 1
            self._result_queue.close()
            self._result_queue.join_thread()
            sys.exit(ret)
          # os.kill(os.getpid(), signal.SIGTERM)



    # --------------------------------------------------------------------------
    #
    def _result_watcher(self):

        try:
            while True:

                try:
                    res = self._result_queue.get(timeout=0.1)
                    self._log.debug('got   result: %s', res)
                    self._result_cb(res)

                except queue.Empty:
                    pass

        except:
            self._log.exception('queue error')
            raise


    # --------------------------------------------------------------------------
    #
    def _result_cb(self, result):

        task, out, err, ret, val, exc = result
        self._log.debug('result cb: task %s', task['uid'])

        with self._plock:
            pid  = task['pid']
            del self._pool[pid]

        # free resources again for the task
        self._dealloc(task)

        task['stdout']           = out
        task['stderr']           = err
        task['exit_code']        = ret
        task['return_value']     = val
        task['exception']        = exc[0]
        task['exception_detail'] = exc[1]

        self._res_put.put(task)
        self._prof.prof('req_stop', uid=task['uid'], msg=self._uid)


    # --------------------------------------------------------------------------
    #
    def _error_cb(self, error):

        self._log.debug('error: %s', error)
        raise RuntimeError(error)


# ------------------------------------------------------------------------------





__copyright__ = 'Copyright 2023, The RADICAL-Cybertools Team'
__license__   = 'MIT'

# pylint: disable=global-statement   # W0603 global `_components`

import os
import time
import signal

import radical.utils   as ru

from ..messages  import ComponentStartedMessage


# ------------------------------------------------------------------------------
#
class ComponentManager(object):
    '''
    RP spans a hierarchy of component instances: the application has a pmgr and
    tmgr, and the tmgr has a staging component and a scheduling component, and
    the pmgr has a launching component, and components also can have bridges,
    etc. This ComponentManager centralises the code needed to spawn, manage and
    terminate such components. Any code which needs to create component should
    create a ComponentManager instance and pass the required component and
    bridge layout and configuration.  Calling `stop()` on the cmgr will
    terminate the components and bridges.
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, sid, reg_addr, owner):

        # create a registry client to obtain the session config and to store
        # component and bridge configs

        self._sid      = sid
        self._reg_addr = reg_addr
        self._owner    = owner
        self._to_kill  = list()

        self._reg    = ru.zmq.RegistryClient(url=self._reg_addr)
        self._cfg    = ru.Config(from_dict=self._reg['cfg'])

        self._uid    = ru.generate_id('cmgr.%(item_counter)04d',
                                      ru.ID_CUSTOM, ns=self._sid)

        self._prof   = ru.Profiler(self._uid, ns='radical.pilot',
                                   path=self._cfg.path)
        self._log    = ru.Logger(self._uid, ns='radical.pilot',
                                 path=self._cfg.path,
                                 level=self._cfg.log_lvl,
                                 debug=self._cfg.debug_lvl)

        self._prof.prof('init2', uid=self._uid, msg=self._cfg.path)

        self._log.debug('cmgr %s (%s)', self._uid, self._owner)

        # component managers open a zmq pipe so that components and bridges can
        # send registration messages.
        self._startups = dict()  # startup messages we have seen

        def register_cb(msg):
            self._log.debug('got message: %s', msg)
            msg = ru.zmq.Message.deserialize(msg)
            if isinstance(msg, ComponentStartedMessage):
                self._startups[msg.uid] = msg
                self._to_kill.append(msg.pid)
            else:
                self._log.error('unknown message type: %s', type(msg))

        self._pipe = ru.zmq.Pipe(mode=ru.zmq.MODE_PULL)
        self._pipe.register_cb(register_cb)
        self._cfg.cmgr_url = str(self._pipe.url)


    # --------------------------------------------------------------------------
    #
    def _wait_startup(self, uids, timeout):
        '''
        Wait for the startup message of the given component UIDs to appear.  If
        that does not happen before timeout, an exception is raised.
        '''

        start = time.time()
        ok    = list()
        nok   = uids
        while True:

            self._log.debug('wait for : %s', nok)

            ok  = [uid for uid in uids if uid     in self._startups]
            nok = [uid for uid in uids if uid not in ok]

            if len(ok) == len(uids):
                break

            if time.time() - start > timeout:
                self._log.debug('wait failed: %s', nok)
                raise RuntimeError('uids %s not found' % nok)

            time.sleep(0.25)

        self._log.debug('wait for done: %s', ok)


    # --------------------------------------------------------------------------
    #
    @property
    def uid(self):
        return self._uid


    # --------------------------------------------------------------------------
    #
    def start_bridges(self, bridges):

        if not bridges:
            return

        self._prof.prof('start_bridges_start', uid=self._uid)

        buids = list()
        for bname, bcfg in bridges.items():

            self._log.debug('start bridge %s', bname)

            uid = bname
            buids.append(uid)

            bcfg.uid       = uid
            bcfg.channel   = bname
            bcfg.cmgr      = self.uid
            bcfg.cmgr_url  = self._cfg.cmgr_url
            bcfg.owner     = self._owner
            bcfg.sid       = self._cfg.sid
            bcfg.path      = self._cfg.path
            bcfg.reg_addr  = self._cfg.reg_addr
            bcfg.log_lvl   = self._cfg.log_lvl
            bcfg.debug_lvl = self._cfg.debug_lvl

            self._reg['bridges.%s.cfg' % bname] = bcfg

            cmd = 'radical-pilot-bridge %s %s %s %s' \
                % (self._sid, self._reg.url, bname, os.getpid())

            out, err, ret = ru.sh_callout(cmd, cwd=self._cfg.path)

            if ret:
                msg = 'bridge startup failed [%s] [%s]', out, err
                self._log.error(msg)
                raise RuntimeError(msg)

            self._log.info('created bridge %s [%s]', bname, bname)

        # all bridges are started, wait for their startup messages
        self._log.debug('wait   for %s', buids)
        self._wait_startup(buids, timeout=10.0)

        self._prof.prof('start_bridges_stop', uid=self._uid)


    # --------------------------------------------------------------------------
    #
    def start_components(self, components, cfg = None):

        if not components:
            return

        self._prof.prof('start_components_start: %s', uid=self._uid)

        cuids = list()
        for cname, ccfg in components.items():

            for _ in range(ccfg.get('count', 1)):

                uid = ru.generate_id(cname + '.%(item_counter)04d',
                                     ru.ID_CUSTOM, ns=self._sid)
                cuids.append(uid)

                ccfg.uid       = uid
                ccfg.kind      = cname
                ccfg.owner     = self._owner
                ccfg.sid       = self._cfg.sid
                ccfg.cmgr      = self._cfg.uid
                ccfg.cmgr_url  = self._cfg.cmgr_url
                ccfg.base      = self._cfg.base
                ccfg.path      = self._cfg.path
                ccfg.reg_addr  = self._cfg.reg_addr
                ccfg.proxy_url = self._cfg.proxy_url
                ccfg.log_lvl   = self._cfg.log_lvl
                ccfg.debug_lvl = self._cfg.debug_lvl

                if cfg:
                    ru.dict_merge(ccfg, cfg, ru.OVERWRITE)

                self._reg['components.%s.cfg' % uid] = ccfg

                self._log.info('create  component %s [%s]', cname, uid)

                cmd = 'radical-pilot-component %s %s %s %s' \
                      % (self._sid, self._reg.url, uid, os.getpid())
                out, err, ret = ru.sh_callout(cmd, cwd=self._cfg.path)

                self._log.debug('component startup out: %s' , out)
                self._log.debug('component startup err: %s' , err)

                if ret:
                    raise RuntimeError('component startup failed')

                self._log.info('created component %s [%s]', cname, uid)

        # all components should start now, wait for heartbeats to appear.
        self._log.debug('wait   for %s', cuids)
        self._wait_startup(cuids, timeout=10.0)

        self._prof.prof('start_components_stop', uid=self._uid)


    # --------------------------------------------------------------------------
    #
    def close(self):

        self._prof.prof('close', uid=self._uid)

        for pid in self._to_kill:

            self._log.debug('kill %s', pid)

            try:
                os.kill(pid, signal.SIGKILL)

            except ProcessLookupError:
                pass

        self._pipe.stop()


# ------------------------------------------------------------------------------





__copyright__ = 'Copyright 2022-2024, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import datetime

# configure the psij logger (captured in the launch components stderr)
import logging
logging.basicConfig(level='DEBUG')

import threading as mt

from .base import PilotLauncherBase
from ...   import states as rps

SCHEMA_LOCAL = 'local'

# psij is optional
psij    = None
psij_ex = None

try:
    import psij
except ImportError as ex:
    psij_ex = ex

OPTIONS_MAPPING = {
    'slurm': {'key': 'constraint',  'delim': '&'},
    'pbs'  : {'key': 'l',           'delim': ','},
    'lsf'  : {'key': 'alloc_flags', 'delim': ' '}
}
SCHEMA_ALIAS = {
    'pbspro': 'pbs'
}


# ------------------------------------------------------------------------------
#
class PilotLauncherPSIJ(PilotLauncherBase):

    # --------------------------------------------------------------------------
    #
    def __init__(self, name, log, prof, state_cb):


        # psij is an optional dependency - let an import exception fall through
        # to disable this pilot launcher
        if psij_ex:
            raise psij_ex

        assert psij

        PilotLauncherBase.__init__(self, name, log, prof, state_cb)

        self._jobs   = dict()      # map pilot id to psi_j job
        self._pilots = dict()      # map psi_j id to pilot job
        self._jex    = dict()      # map launch schema to psi_j job executors
        self._lock   = mt.RLock()  # lock above structures


    # --------------------------------------------------------------------------
    #
    def _get_schema(self, rcfg):

        url     = rcfg['job_manager_endpoint']
        schemas = url.split(':')[0].split('+')

        if 'ssh' in schemas:
            schemas.remove('ssh')

        if 'gsissh' in schemas:
            schemas.remove('gsissh')

        if len(schemas) > 1:
            return

        if not schemas:
            return

        schema = schemas[0]

        if schema in SCHEMA_ALIAS:
            schema = SCHEMA_ALIAS[schema]

        if schema == 'fork':
            schema = SCHEMA_LOCAL

        return schema


    # --------------------------------------------------------------------------
    #
    def _translate_state(self, status):

        if   status.state == psij.JobState.NEW       : return rps.NEW
        elif status.state == psij.JobState.QUEUED    : return rps.PMGR_LAUNCHING
        elif status.state == psij.JobState.ACTIVE    : return rps.PMGR_ACTIVE
        elif status.state == psij.JobState.COMPLETED : return rps.DONE
        elif status.state == psij.JobState.FAILED    : return rps.FAILED
        elif status.state == psij.JobState.CANCELED  : return rps.CANCELED
        else:
            raise ValueError('cannot interpret psij state: %s' % repr(status))


    # --------------------------------------------------------------------------
    #
    def _job_status_cb(self, job, status):

        try:
            with self._lock:

                if job.id not in self._pilots:
                    return

                rp_state = self._translate_state(status)
                pilot    = self._pilots[job.id]

            # we don't report PMGR+ACTIVE here - that information comes from the
            # pilot itself once it is up and running
            if rp_state != rps.PMGR_ACTIVE:
                self._state_cb(pilot, rp_state)

        except Exception:
            self._log.exception('job status callback failed')


    # --------------------------------------------------------------------------
    #
    def can_launch(self, rcfg, pilot):

        schema = self._get_schema(rcfg)

        if not schema:
            return False

        if schema not in self._jex:

            self._log.debug('create executor for %s', schema)
            try:
                self._jex[schema] = psij.JobExecutor.get_instance(schema)
                self._jex[schema].set_job_status_callback(self._job_status_cb)
            except:
                self._log.exception('failed to create psij executor')
                return False

        return True


    # --------------------------------------------------------------------------
    #
    def launch_pilots(self, rcfg, pilots):

        assert psij

        for pilot in pilots:

            pid    = pilot['uid']
            schema = self._get_schema(rcfg)
            assert schema

            jex = self._jex.get(schema)

            assert jex

            jd = pilot['jd_dict']

            if jd.project and ':' in jd.project:
                proj, res = jd.project.split(':', 1)
            else:
                proj, res = jd.project, None

            if jd.queue and ':' in jd.queue:
                part, qos = jd.queue.split(':', 1)
            else:
                part, qos = jd.queue, None

            attr = psij.JobAttributes()
            attr.duration       = datetime.timedelta(minutes=jd.wall_time_limit)
            attr.queue_name     = part
            attr.project_name   = proj
            attr.reservation_id = res

            if qos:
                attr.set_custom_attribute(name=f'{schema}.qos', value=qos)

            # define batch options/constraints
            if (jd.system_architecture.get('options') and
                    schema in OPTIONS_MAPPING):
                opts = OPTIONS_MAPPING[schema]
                attr.set_custom_attribute(
                    name=f'{schema}.{opts["key"]}',
                    value=opts['delim'].join(jd.system_architecture['options']))

            spec = psij.JobSpec()
            spec.attributes          = attr
            spec.executable          = jd.executable
            spec.arguments           = jd.arguments
            spec.environment         = jd.environment
            spec.directory           = jd.working_directory
            spec.stdout_path         = jd.output
            spec.stderr_path         = jd.error
            # inherit environment for local executor only
            spec.inherit_environment = bool(schema == SCHEMA_LOCAL)

            spec.resources = psij.ResourceSpecV1(
                node_count=jd.node_count or None,
                process_count=jd.total_cpu_count,
                exclusive_node_use=bool(jd.system_architecture.get('exclusive'))
            )
          # spec.resources.cpu_cores_per_process = 1
          # spec.resources.gpu_cores_per_process = jd.total_gpu_count

            job = psij.Job(spec)

            self._jobs[pid]      = job
            self._pilots[job.id] = pilot
            self._log.debug('added %s: %s', job.id, pid)

            jex.submit(job)


    # --------------------------------------------------------------------------
    #
    def kill_pilots(self, pids):

        for pid in pids:
            if pid not in pids:
                continue

            self._jobs[pid].cancel()


# ------------------------------------------------------------------------------




#!/usr/bin/env python

__copyright__ = 'Copyright 2013-2014, http://radical.rutgers.edu'
__license__   = 'MIT'

import os
import sys

import radical.pilot as rp
import radical.utils as ru


pwd = os.path.dirname(os.path.abspath(__file__))
dh  = ru.DebugHelper()


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    # we use a reporter class for nicer output
    report = ru.Reporter(name='radical.pilot')
    report.title('Getting Started (RP version %s)' % rp.version)

    # use the resource specified as argument, fall back to localhost
    if   len(sys.argv)  > 2: report.exit('Usage:\t%s [resource]\n\n' % sys.argv[0])
    elif len(sys.argv) == 2: resource = sys.argv[1]
    else                   : resource = 'local.localhost'

    # Create a new session. No need to try/except this: if session creation
    # fails, there is not much we can do anyways...
    session = rp.Session()

    # all other pilot code is now tried/excepted.  If an exception is caught, we
    # can rely on the session object to exist and be valid, and we can thus tear
    # the whole RP stack down via a 'session.close()' call in the 'finally'
    # clause...
    try:

        # read the config used for resource details
        report.info('read config')
        config = ru.read_json('%s/../config.json' % pwd)
        report.ok('>>ok\n')

        report.header('submit pilots')

        # Add a Pilot Manager. Pilot managers manage one or more Pilots.
        pmgr = rp.PilotManager(session=session)

        # Define an [n]-core local pilot that runs for [x] minutes
        # Here we use a dict to initialize the description object
        pd_init = {'resource'      : resource,
                   'runtime'       : 15,  # pilot runtime (min)
                   'exit_on_error' : True,
                   'project'       : config[resource]['project'],
                   'queue'         : config[resource]['queue'],
                   'access_schema' : config[resource]['schema'],
                   'cores'         : config[resource]['cores']
                  }
        pdesc = rp.PilotDescription(pd_init)

        # Launch the pilot.
        pilot = pmgr.submit_pilots(pdesc)

        report.header('submit tasks')

        # Register the Pilot in a TaskManager object.
        tmgr = rp.TaskManager(session=session)
        tmgr.add_pilots(pilot)

        # Create a workload of Tasks.
        # Each compute task runs '/bin/date'.

        n = 32  # number of tasks to run
        report.info('create %d task description(s)\n\t' % n)

        tds = list()
        for i in range(0, n):

            # create a new CU description, and fill it.
            # Here we don't use dict initialization.
            td = rp.TaskDescription()
            td.executable       = '/bin/sleep'
            td.arguments        = ['60']
            td.gpu_processes    = 0
            td.cpu_processes    = 2
            td.cpu_threads      = 2
          # td.cpu_process_type = rp.MPI
          # td.cpu_thread_type  = rp.OpenMP
            tds.append(td)
            report.progress()
        report.ok('>>ok\n')

        # Submit the previously created Task descriptions to the
        # PilotManager. This will trigger the selected scheduler to start
        # assigning Tasks to the Pilots.
        tasks = tmgr.submit_tasks(tds)

        # die - this is where the restart attempt will continue
        # but before, we store some UIDs
        with ru.ru_open('%s/restart.dat' % pwd, 'w') as fout:
            fout.write('session_id %s\n' % session.uid)
            fout.write('tmgr_ids   %s\n' % tmgr.uid)
            fout.write('pmgrs_ids  %s\n' % pmgr.uid)
            fout.write('pilot_ids  %s\n' % pilot.uid)
            fout.write('task_ids   %s\n' % ':'.join(u.uid for u in tasks))

        pilot.wait(state=rp.PMGR_ACTIVE)
        print('killme %s' % os.getpid())
        sys.exit(0)


    except Exception as e:
        # Something unexpected happened in the pilot code above
        report.error('caught Exception: %s\n' % e)
        ru.print_exception_trace()
        raise

    except (KeyboardInterrupt, SystemExit):
        # the callback called sys.exit(), and we can here catch the
        # corresponding KeyboardInterrupt exception for shutdown.  We also catch
        # SystemExit (which gets raised if the main threads exits for some other
        # reason).
        ru.print_exception_trace()
        report.warn('exit requested\n')

    finally:
        # always clean up the session, no matter if we caught an exception or
        # not.  This will kill all remaining pilots.
        report.header('finalize')
      # session.close(download=False)

    report.header()


# ------------------------------------------------------------------------------





__copyright__ = 'Copyright 2016-2022, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import os

import radical.utils as ru

from .base import LaunchMethod


# ------------------------------------------------------------------------------
#
# aprun: job launcher for Cray systems (alps-run)
# TODO : ensure that only one concurrent aprun per node is executed!
#
class APRun(LaunchMethod):

    # --------------------------------------------------------------------------
    #
    def __init__(self, name, lm_cfg, rm_info, log, prof):

        self._command: str = ''

        LaunchMethod.__init__(self, name, lm_cfg, rm_info, log, prof)


    # --------------------------------------------------------------------------
    #
    def _init_from_scratch(self, env, env_sh):

        lm_info = {'env'    : env,
                   'env_sh' : env_sh,
                   'command': ru.which('aprun')}

        return lm_info


    # --------------------------------------------------------------------------
    #
    def _init_from_info(self, lm_info):

        self._env         = lm_info['env']
        self._env_sh      = lm_info['env_sh']
        self._command     = lm_info['command']

        assert self._command


    # --------------------------------------------------------------------------
    #
    def finalize(self):

        pass


    # --------------------------------------------------------------------------
    #
    def can_launch(self, task):

        if not task['description']['executable']:
            return False, 'no executable'

        return True, ''


    # --------------------------------------------------------------------------
    #
    def get_launcher_env(self):

        return ['. $RP_PILOT_SANDBOX/%s' % self._env_sh]


    # --------------------------------------------------------------------------
    #
    def get_launch_cmds(self, task, exec_path):

        td             = task['description']
        ranks          = td['ranks']
        cores_per_rank = td.get('cores_per_rank', 1)

        # aprun options
        # –  Number of MPI ranks per node:                –N <n_ranks_per_node>
        # –  Total number of MPI ranks:                   –n <n_ranks_total>
        # –  Number of hyperthreads per MPI rank (depth): –d <n_rank_threads>
        # –  Number of hyperthreads per core:             –j <n_hwthreads>
        # –  MPI rank and thread placement:               --cc depth
        # –  Environment variables:                       -e <env_var>
        # –  Core specialization:                         -r <n_threads>

        cmd_options = '-n %s ' % ranks + \
                      '-d %s'  % cores_per_rank

        # CPU affinity binding
        # - use –d and --cc depth to let ALPS control affinity
        # - use --cc none if you want to use OpenMP (or KMP) env. variables
        #   to specify affinity: --cc none -e KMP_AFFINITY=<affinity>
        #   (*) turn off thread affinity: export KMP_AFFINITY=none
        #
        # saga_smt = os.environ.get('RADICAL_SAGA_SMT')
        # if saga_smt:
        #     cmd_options += ' -j %s' % saga_smt
        #     cmd_options += ' --cc depth'

        # `share` mode access restricts the application specific cpuset
        # contents to only the application reserved cores and memory on NUMA
        # node boundaries, meaning the application will not have access to
        # cores and memory on other NUMA nodes on that compute node.
        #
        # slots = task['slots']
        # nodes = set([slot['node_name'] for slot in slots])
        # if len(nodes) < 2:
        #     cmd_options += ' -F share'  # default is `exclusive`
        # cmd_options += ' -L %s ' % ','.join(nodes)

        # task_env = td['environment']
        # cmd_options += ''.join([' -e %s=%s' % x for x in task_env.items()])
        # if td['cores_per_rank'] > 1 and 'OMP_NUM_THREADS' not in task_env:
        #     cmd_options += ' -e OMP_NUM_THREADS=%(cores_per_rank)s' % td

        cmd = '%s %s %s' % (self._command, cmd_options, exec_path)
        return cmd.rstrip()


    # --------------------------------------------------------------------------
    #
    def get_rank_cmd(self):

        ret  = 'test -z "$MPI_RANK"    || export RP_RANK=$MPI_RANK\n'
        ret += 'test -z "$PMIX_RANK"   || export RP_RANK=$PMIX_RANK\n'
        ret += 'test -z "$ALPS_APP_PE" || export RP_RANK=$ALPS_APP_PE\n'

        return ret


# ------------------------------------------------------------------------------





__copyright__ = 'Copyright 2014-2022, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import os
import json
import stat
import time
import socket

import threading           as mt

import radical.utils       as ru

from ..   import utils     as rpu
from ..   import states    as rps
from ..   import constants as rpc
from ..   import Session
from ..   import TaskDescription, AGENT_SERVICE, TASK_SERVICE

from ..resource_config import RO


# ------------------------------------------------------------------------------
#
class Agent_0(rpu.AgentComponent):

    '''
    This is the main agent.  It starts sub-agents and watches them.  If any of
    the sub-agents die, it will shut down the other sub-agents and itself.

    This class inherits the rpu.AgentComponent, so that it can use its
    communication bridges and callback mechanisms.
    '''

    _shell = ru.which('bash') or '/bin/sh'


    # --------------------------------------------------------------------------
    #
    def __init__(self):

        cfg = ru.Config(path='./agent_0.cfg')

        self._uid     = cfg.uid
        self._pid     = cfg.pid
        self._sid     = cfg.sid
        self._owner   = cfg.owner
        self._pmgr    = cfg.pmgr
        self._pwd     = cfg.pilot_sandbox

        self._session = Session(uid=cfg.sid, cfg=cfg, _role=Session._AGENT_0)

        self._rm      = self._session.get_rm()

        # init the worker / component base classes, connects registry
        super().__init__(cfg, self._session)

        self._starttime   = time.time()
        self._final_cause = None

        # keep some state about service startups
        self._service_uid_launched = None
        self._service_uids_running = list()
        self._service_start_evt    = mt.Event()
        self._service_lock         = mt.Lock()  # launch one service at a time

        # this is the earliest point to sync bootstrap and agent profiles
        self._prof.prof('hostname', uid=cfg.pid, msg=ru.get_hostname())

        # ensure that app communication channels are visible to workload
        self._configure_app_comm()

        # start the sub agents
        self._start_sub_agents()

        # regularly check for lifetime limit
        self.register_timed_cb(self._check_lifetime, timer=10)

        # also open a service endpoint so that a ZMQ client can submit tasks to
        # this agent
        self._service = None
        self._start_service_ep()


    # --------------------------------------------------------------------------
    #
    def _proxy_input_cb(self, msg):

        self._log.debug_8('proxy input cb: %s', len(msg))

        services   = list()
        to_advance = list()

        for task in msg:

            # make sure the tasks obtain env settings (if needed)
            if 'task_environment' in self.session.rcfg:

                if not task['description'].get('environment'):
                    task['description']['environment'] = dict()

                for k,v in self.session.rcfg.task_environment.items():
                    # FIXME: this might overwrite user specified env
                    task['description']['environment'][k] = v

            # FIXME: raise or fail task!
            if task['state'] != rps.AGENT_STAGING_INPUT_PENDING:
                self._log.error('invalid state: %s:%s:%s', task['uid'],
                        task['state'], task.get('states'))
                continue

            if task['description'].get('mode') == TASK_SERVICE:
                services.append(task)

            else:
                to_advance.append(task)

        # now we really own the tasks and can start working on them (ie. push
        # them into the pipeline).  We don't publish nor profile as advance,
        # since the state transition happened already on the client side when
        # the state was set.
        self.advance(to_advance, publish=False, push=True)

        for service in services:
            self._launch_service_task(service)


    # --------------------------------------------------------------------------
    #
    def _proxy_output_cb(self, msg):

        # we just forward the tasks to the task proxy queue
        self._log.debug('proxy output cb: %s', len(msg))
        self.advance(msg, publish=False, push=True, qname=self._sid)


    # --------------------------------------------------------------------------
    #
    def _client_ctrl_cb(self, topic, msg):

        self._log.debug('ctl sub cb: %s %s', topic, msg)
        ## FIXME?


    # --------------------------------------------------------------------------
    #
    def _configure_app_comm(self):

        # if the pilot description contains a request for application comm
        # channels, merge those into the agent config
        #
        # FIXME: this needs to start the app_comm bridges
        app_comm = self.session.rcfg.get('app_comm')
        if app_comm:

            # bridge addresses also need to be exposed to the workload
            if 'task_environment' not in self.session.rcfg:
                self.session.rcfg['task_environment'] = dict()

            if isinstance(app_comm, list):
                app_comm = {ac: {'bulk_size': 0,
                                 'stall_hwm': 1,
                                 'log_level': 'error'} for ac in app_comm}
            for ac in app_comm:

                if ac in self._reg['bridges']:
                    raise ValueError('reserved app_comm name %s' % ac)

                self._reg['bridges.%s' % ac] = app_comm[ac]

                AC = ac.upper()

                self.session.rcfg.task_environment['RP_%s_IN'  % AC] = ac['addr_in']
                self.session.rcfg.task_environment['RP_%s_OUT' % AC] = ac['addr_out']


    # --------------------------------------------------------------------------
    #
    def initialize(self):

        # we  first register outputs, then start known service tasks, and only
        # once those are up and running we register inputs.  This ensures that
        # we do not start any incoming tasks before we services are up.

        # forward received tasks to agent input staging
        self.register_output(rps.AGENT_STAGING_INPUT_PENDING,
                             rpc.AGENT_STAGING_INPUT_QUEUE)

        # and return completed tasks to the task manager
        self.register_output(rps.TMGR_STAGING_OUTPUT_PENDING,
                             rpc.PROXY_TASK_QUEUE)

        # before we run any tasks (including service tasks), prepare a named_env
        # `rp` for tasks which use the pilot's own environment, such as raptors
        #
        # register the respective RPC handler...
        self.register_rpc_handler('prepare_env', self._prepare_env,
                                                 rpc_addr=self._pid)

        # ...prepare the pilot's own env...
        env_spec = {'type'    : os.environ['RP_VENV_TYPE'],
                    'path'    : os.environ['RP_VENV_PATH'],
                    'pre_exec': ['export PYTHONPATH=%s'
                                 %  os.environ.get('PYTHONPATH', ''),
                                 'export PATH=%s'
                                 %  os.environ.get('PATH', '')]
                   }
        self.rpc('prepare_env', env_name='rp', env_spec=env_spec,
                                rpc_addr=self._pid)

        # ...and prepare all envs which are defined in the pilot description
        for env_name, env_spec in self._cfg.get('prepare_env', {}).items():
            self.rpc('prepare_env', env_name=env_name, env_spec=env_spec,
                                    rpc_addr=self._pid)

        # launch predefined services
        for sd in self._cfg.services:
            self._launch_service(sd)

        # allow registration of external services
        self.register_rpc_handler('register_service', self._register_service,
                                                      rpc_addr=self._pid)

        # listen for new tasks from the client
        self.register_input(rps.AGENT_STAGING_INPUT_PENDING,
                            rpc.PROXY_TASK_QUEUE,
                            qname=self._pid,
                            cb=self._proxy_input_cb)

        # listen for completed tasks to forward to client
        self.register_input(rps.TMGR_STAGING_OUTPUT_PENDING,
                            rpc.AGENT_COLLECTING_QUEUE,
                            cb=self._proxy_output_cb)

        # sub-agents are started, components are started, bridges are up: we are
        # ready to roll!  Send state update
        rm_info = self._rm.info
        n_nodes = len(rm_info['node_list'])

        self._log.debug('advance to PMGR_ACTIVE')

        rest_url = None
        if self._service:
            rest_url = self._service.addr

        pilot = {'$all'     : True,              # pass full info to client side
                 'type'     : 'pilot',
                 'uid'      : self._pid,
                 'state'    : rps.PMGR_ACTIVE,
                 'rest_url' : rest_url,
                 'resources': {'rm_info': rm_info,
                               'cpu'    : rm_info['cores_per_node'] * n_nodes,
                               'gpu'    : rm_info['gpus_per_node']  * n_nodes}}

        self.advance(pilot, publish=True, push=False, fwd=True)


    # --------------------------------------------------------------------------
    #
    def work(self):

        # all work is done in the registered callbacks
        time.sleep(1)


    # --------------------------------------------------------------------------
    #
    def stage_output(self):

        if  os.path.isfile('./staging_output.txt'):

            if not os.path.isfile('./staging_output.tgz'):

                cmd = 'tar zcvf staging_output.tgz $(cat staging_output.txt)'
                out, err, ret = ru.sh_callout(cmd, shell=True)

                if ret:
                    self._log.debug('out: %s', out)
                    self._log.debug('err: %s', err)
                    self._log.error('output tarring failed: %s', cmd)


    # --------------------------------------------------------------------------
    #
    def finalize(self):

        # tar up output staging data
        self._log.debug('stage output parent')
        self.stage_output()

        self._log.info('rusage: %s', rpu.get_rusage())

        out, err, log = '', '', ''

        try   : out = ru.ru_open('./agent_0.out', 'r').read(1024)
        except: pass
        try   : err = ru.ru_open('./agent_0.err', 'r').read(1024)
        except: pass
        try   : log = ru.ru_open('./agent_0.log', 'r').read(1024)
        except: pass

        self._log.debug('final cause: %s', self._final_cause)

        if   self._final_cause == 'timeout'  : state = rps.DONE
        elif self._final_cause == 'cancel'   : state = rps.CANCELED
        elif self._final_cause == 'sys.exit' : state = rps.CANCELED
        else                                 : state = rps.FAILED

        # NOTE: we do not push the final pilot state, as that is done by the
        #       bootstrapper *after* this pilot *actually* finished.
        with ru.ru_open('./killme.signal', 'w') as fout:
            fout.write('%s\n' % state)

        pilot = {'type'   : 'pilot',
                 'uid'    : self._pid,
                 'stdout' : out,
                 'stderr' : err,
                 'logfile': log,
                 'state'  : state}

        self._log.debug('push final state update')
        self._log.debug('update state: %s: %s', state, self._final_cause)
        self.advance(pilot, publish=True, push=False)

        # tear things down in reverse order
        self._rm.stop()
        self._session.close()


    # --------------------------------------------------------------------------
    #
    def _launch_service(self, td_orig):

        td = TaskDescription(td_orig)

        if not td.uid:
            td.uid = ru.generate_id('service.%(item_counter)04d',
                                    ru.ID_CUSTOM, ns=self.session.uid)

        if not td_orig.get('name')    : td.name     = td.uid
        if not td_orig.get('mode')    : td.mode     = AGENT_SERVICE
        if not td_orig.get('metadata'): td.metadata = dict()

        td.verify()

        self._log.info('starting agent service from sd %s', td.uid)

        sbox = self._cfg.pilot_sandbox + '/' + td.uid

        task = dict()
        task['uid']               = td.uid
        task['name']              = td.name or td.uid
        task['origin']            = 'agent'
        task['type']              = 'task'
        task['pilot']             = self._cfg.pid
        task['description']       = td.as_dict()
        task['pilot_sandbox']     = self._cfg.pilot_sandbox
        task['session_sandbox']   = self._cfg.session_sandbox
        task['resource_sandbox']  = self._cfg.resource_sandbox
        task['task_sandbox']      = 'file://localhost/' + sbox
        task['task_sandbox_path'] = sbox
        task['resources']         = {'cpu': td.ranks * td.cores_per_rank,
                                     'gpu': td.ranks * td.gpus_per_rank}

        self._launch_service_task(task)


    # ----------------------------------------------------------------------
    #
    def _launch_service_task(self, task):

        tid = task['uid']

        self._log.info('starting agent service %s', tid)

        td = TaskDescription(task['description'])

        if not task['description'].get('mode')    : td.mode     = AGENT_SERVICE
        if not task['description'].get('metadata'): td.metadata = dict()

        # we wrap the service
        exe, args, pat = td.executable, td.arguments, td.info_pattern

        td.executable = 'radical-pilot-service-wrapper'
        td.arguments  = ['-c', '%s %s' % (exe, ' '.join(args))]
        td.arguments += ['-u', tid]
        td.arguments += ['-P', os.getpid()]
        td.arguments += ['-v']

        if td.stdout:
            td.arguments += ['-o', td.stdout]

        if td.stderr:
            td.arguments += ['-e', td.stderr]

        if td.startup_timeout:
            td.arguments += ['-t', '%d' % td.startup_timeout]

        if pat:
            pat_src, pat_regex = pat.split(':', 1)
            td.arguments += ['-m', pat_src, '-p', pat_regex]

        # ensure that the description is viable
        td.verify()

        task['description'] = td.as_dict()
        task['state']       = rps.AGENT_STAGING_INPUT_PENDING

        with self._service_lock:

            self._service_start_evt.clear()

            self._log.debug('set agent service id to %s', tid)
            self._service_uid_launched = tid

            self.advance(task, publish=False, push=True)

            # At this point we wait for one of two events to happen: either the
            # task will go into a state beyond `AGENT_EXECUTING` and will thus
            # have completed prematurely in which case we declare a failure, or
            # the service comes up and the wrapper sends a control message.  In
            # the former case, the state cb will trigger the service startup
            # event, in the latter case, that message will trigger that event as
            # well.
            #
            # Or we time out of course :-)
            #
            # FIXME: need to watch state updates

            # The service task timeout is watched by the wrapper, but we also
            # watch the wrapper itself.
            if td.startup_timeout:
                if not self._service_start_evt.wait(timeout=td.startup_timeout):
                    raise RuntimeError('Unable to start service')
            else:
                self._service_start_evt.wait()

            info = self._reg.get('services.%s' % td.uid)
            self._log.info('agent service started: %s - %s', td.uid, info)

            # send a notification, specifically also to the client side
            if not info:
                info = 'service is up'

            self.publish(rpc.CONTROL_PUBSUB, {'cmd' : 'service_up',
                                              'arg' : {'uid' : td.uid,
                                                       'info': info},
                                              'fwd' : True})


    # --------------------------------------------------------------------------
    #
    def _start_sub_agents(self):
        '''
        For the list of sub_agents, get a launch command and launch that
        agent instance on the respective node.  We pass it to the seconds
        bootstrap level, there is no need to pass the first one again.
        '''

        # FIXME: reroute to agent daemonizer

        if not self._cfg.agents:
            return

        n_agents      = len(self._cfg.agents)
        n_agent_nodes = len(self._rm.info.agent_node_list)

        assert n_agent_nodes >= n_agents


        self._log.debug('start_sub_agents')

        # store the current environment as the sub-agents will use the same
        # (it will be called within "bootstrap_2.sh")
        ru.env_prep(os.environ, script_path='./env/agent.env')

        # the configs are written, and the sub-agents can be started.  To know
        # how to do that we create the agent launch method, have it creating
        # the respective command lines per agent instance, and run via popen.

        bs_name = '%s/bootstrap_2.sh'

        for idx, sa in enumerate(self._cfg.agents):

            target  = self._cfg.agents[sa]['target']
            bs_args = [self._sid, self.session.cfg.reg_addr, sa]

            if target not in ['local', 'node']:

                raise ValueError('agent target unknown (%s)' % target)

            if target == 'local':

                # start agent locally
                bs_path  = bs_name % self._pwd
                cmdline  = self._shell
                cmdline += ' -l %s' % ' '.join([bs_path] + bs_args)

            else:  # target == 'node':

                node = self._rm.info.agent_node_list[idx]
                # start agent remotely, use launch method
                # NOTE:  there is some implicit assumption that we can use
                #        the 'agent_node' string as 'agent_string:0' and
                #        obtain a well format slot...
                # FIXME: it is actually tricky to translate the agent_node
                #        into a viable 'slots' structure, as that is
                #        usually done by the schedulers.  So we leave that
                #        out for the moment, which will make this unable to
                #        work with a number of launch methods.  Can the
                #        offset computation be moved to the ResourceManager?

                launch_script = '%s/%s.launch.sh'   % (self._pwd, sa)
                exec_script   = '%s/%s.exec.sh'     % (self._pwd, sa)

                node_cores = [RO(index=cid, occuapation=rpc.BUSY)
                                            for cid, cstate
                                            in  enumerate(node['cores'])
                                            if  cstate == rpc.FREE]

                node_gpus  = [RO(index=gid, occuapation=rpc.BUSY)
                                            for gid, gstate
                                            in  enumerate(node['gpus'])
                                            if  gstate == rpc.FREE]

                agent_task = {
                    'uid'               : sa,
                    'task_sandbox_path' : self._pwd,
                    'description'       : TaskDescription({
                        'uid'           : sa,
                        'ranks'         : 1,
                        'cores_per_rank': self._rm.info.cores_per_node,
                        'executable'    : self._shell,
                        'arguments'     : [bs_name % self._pwd] + bs_args
                    }).as_dict(),
                    'slots'             : [{'node_name' : node['name'],
                                            'node_index': node['index'],
                                            'cores'     : node_cores,
                                            'gpus'      : node_gpus,
                                            'lfs'       : 0,
                                            'mem'       : 0}]
                }

                # find a launcher to use
                launcher, lname = self._rm.find_launcher(agent_task)
                if not launcher:
                    raise RuntimeError('no launch method found for sub agent')

                self._log.debug('found sa launcher %s', lname)

                # FIXME: set RP environment (as in Popen Executor)

                tmp  = '#!%s\n\n' % self._shell
                tmp += 'export RP_PILOT_SANDBOX="%s"\n\n' % self._pwd
                cmds = launcher.get_launcher_env()
                for cmd in cmds:
                    tmp += '%s || exit 1\n' % cmd

                cmds = launcher.get_launch_cmds(agent_task, exec_script)
                tmp += '%s\nexit $?\n\n' % cmds
                with ru.ru_open(launch_script, 'w') as fout:
                    fout.write(tmp)

                tmp  = '#!%s\n\n' % self._shell
                tmp += self._shell
                tmp += ' -l %s\n\n' % ' '.join([bs_name % '.'] + bs_args)
                with ru.ru_open(exec_script, 'w') as fout:
                    fout.write(tmp)

                # make sure scripts are executable
                st_l = os.stat(launch_script)
                st_e = os.stat(exec_script)
                os.chmod(launch_script, st_l.st_mode | stat.S_IEXEC)
                os.chmod(exec_script,   st_e.st_mode | stat.S_IEXEC)

                # spawn the sub-agent
                cmdline = launch_script

            self._log.info('create sub-agent %s: %s', sa, cmdline)
            ru.sh_callout_bg(cmdline, stdout='%s.out' % sa,
                                      stderr='%s.err' % sa,
                                      cwd=self._pwd)

        self._log.debug('start_sub_agents done')


    # --------------------------------------------------------------------------
    #
    def _check_lifetime(self):

        # Make sure that we haven't exceeded the runtime - otherwise terminate.
        if self._cfg.runtime:

            if time.time() >= self._starttime + (int(self._cfg.runtime) * 60):

                self._log.info('runtime limit (%ss).', self._cfg.runtime * 60)
                self._final_cause = 'timeout'
                self.stop()
                return False  # we are done

        return True


    # --------------------------------------------------------------------------
    #
    def control_cb(self, topic, msg):
        '''
        Check for commands on the control pubsub, mainly waiting for RPC
        requests to handle.
        '''

        self._log.debug_1('control msg %s: %s', topic, msg)

        cmd = msg.get('cmd')
        arg = msg.get('arg')

        self._log.debug('pilot command: %s: %s', cmd, arg)
        self._prof.prof('cmd', msg="%s : %s" %  (cmd, arg), uid=self._pid)

        if cmd == 'pmgr_heartbeat' and arg['pmgr'] == self._pmgr:
            self._session._hb.beat(uid=self._pmgr)
            return True

        elif cmd == 'cancel_pilots':
            return self._ctrl_cancel_pilots(msg)

        elif cmd == 'service_info':
            self._log.debug('PILOT COMMAND: %s: %s', cmd, arg)
            return self._ctrl_service_info(msg, arg)

        else:
            self._log.warn('invalid command: [%s]', cmd)


    # --------------------------------------------------------------------------
    #
    def command_port(self):
        '''
        listen on a command port, relay incoming commands to the control pubsub
        '''

        sock = None
        while True:
            port = ru.find_port()
            try:
                # we only listen on all interfaces
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                hostip = ru.get_hostip()
                sock.bind((hostip, port))
                break
            except Exception as e:
                self._log.error('port %s unusable: %s', port, e)
                pass

        assert sock

        self._log.debug('bound cmd handler to port %s', port)

        ru.write_json('agent_0.ports.json', {'command_port'  : port,
                                             'command_hostip': hostip})
        sock.listen(1)

        while True:
            connection, client_address = sock.accept()

            try:
                self._log.info('command connection from %s', client_address)

                data = connection.recv(1024)
                msg  = json.loads(data)

                # don't forward to the client
                if 'fwd' not in msg:
                    msg['fwd'] = False

                self._log.debug('command: %s', msg)

                self.publish(rpc.CONTROL_PUBSUB, msg)
                connection.sendall(b'OK')

            except Exception as e:
                connection.sendall(('ERROR: %s' % e).encode('utf-8'))

            finally:
                connection.close()



    # --------------------------------------------------------------------------
    #
    def stop(self):

        self._log.info('stop agent')
        self._final_cause = 'cancel'
        super().stop()
        self._session.close()


    # --------------------------------------------------------------------------
    #
    def _ctrl_cancel_pilots(self, msg):

        arg = msg['arg']

        if self._pid not in arg.get('uids'):
            self._log.debug('ignore cancel %s', msg)
            return True

        self._log.info('cancel pilot cmd')
        self._final_cause = 'cancel'
        self.publish(rpc.CONTROL_PUBSUB, {'cmd' : 'terminate',
                                          'arg' : None})
        self.stop()

        # work is done - unregister this cb
        return False


    # --------------------------------------------------------------------------
    #
    def _ctrl_service_info(self, msg, arg):

        uid   = arg['uid']
        error = arg['error']
        info  = arg['info']

        self._log.debug('service info: %s: %s', uid, info)

        # This message signals that an agent service instance is up and running.
        # We expect to find the service UID in args and can then unblock the
        # service startup wait for that uid

        if uid != self._service_uid_launched:
            # we do not know this service instance
            self._log.warn('ignore service startup signal for %s [%s]', uid,
                           self._service_uid_launched)
            return True

        if uid in self._service_uids_running:
            self._log.warn('duplicated service startup signal for %s', uid)
            return True

        if error is not None:
            self._log.error('service %s failed: %s', uid, error)
            return True

        self._service_uids_running.append(uid)
        self._log.debug('service %s started (%s): %s', uid,
                        len(self._service_uids_running), info)

        # add info to registry (might be empty!)
        self._reg['services.%s' % uid] = info

        # signal main thread when that the service is up
        self._log.debug('set service start event for %s', uid)
        self._service_start_evt.set()

        return True


    # --------------------------------------------------------------------------
    #
    def _register_service(self, uid, info):

        # add info to registry (might be empty!)
        self._reg['services.%s' % uid] = info

        return True


    # --------------------------------------------------------------------------
    #
    def _prepare_env(self, env_name, env_spec):

        self._log.debug('env_spec %s: %s', env_name, env_spec)

        etype = env_spec.get('type', 'venv')
        evers = env_spec.get('version')
        path  = env_spec.get('path')
        emods = env_spec.get('setup')    or []
        pre   = env_spec.get('pre_exec') or []
        out   = None

        # always load the original pilot env first.
        pre_exec = '-P ". env/bs0_pre_0.sh" '
        for cmd in pre:
            pre_exec += '-P "%s" ' % cmd

        if emods: mods = '-m "%s"' % ','.join(emods)
        else    : mods = ''

      # assert etype == 'virtualenv'
      # assert evers

        # only create a new VE if path is not set or if it does not exist
        if path:
            path = path.rstrip('/')

        ve_local_path = '%s/env/rp_named_env.%s' % (self._pwd, env_name)
        if path: ve_path = path
        else   : ve_path = ve_local_path

        if evers:
            evers = '-v %s' % evers
        else:
            evers = ''

        rp_cse = ru.which('radical-pilot-create-static-ve')
        ve_cmd = '/bin/bash %s -d -p %s -t %s ' % (rp_cse, ve_path, etype) + \
                 '%s %s %s '                    % (evers, mods, pre_exec)  + \
                 '-T %s.env'                    % (ve_local_path)

        # FIXME: we should export all sandboxes etc. to the prep_env.
        os.environ['RP_RESOURCE_SANDBOX'] = '../../'

        # `shell=False` to avoid expansion of env variables in pre_exec lines
        self._log.debug('env cmd: %s', ve_cmd)
        out, err, ret = ru.sh_callout(ve_cmd, shell=False)
        self._log.debug('    out: %s', out)
        self._log.debug('    err: %s', err)
        self._log.debug('    ret: %s', ret)

        if ret:
            raise RuntimeError('prepare_env failed: \n%s\n%s\n' % (out, err))

        # if the ve lives outside of the pilot sandbox, link it
        if path:
            os.symlink(path, ve_local_path)

        self._log.debug('ve_path: %s', ve_path)

        # prepare the env to be loaded in task exec scripts
        with ru.ru_open('%s.sh' % ve_local_path, 'w') as fout:
            if etype == 'shell':
                fout.write('\n. %s.env\n\n' % ve_local_path)
            else:
                fout.write('\n. %s/bin/activate\n\n' % ve_path)

        # publish the venv creation to the scheduler
        self.publish(rpc.CONTROL_PUBSUB, {'cmd': 'register_named_env',
                                          'arg': {'env_name': env_name}})
        return out


    # --------------------------------------------------------------------------
    #
    def _start_service_ep(self):

        if self._cfg.enable_ep:

            self._service = ru.zmq.Server(uid='%s.server' % self._uid)
            self._service.register_request('submit_tasks', self._ep_submit_tasks)
            self._service.start()

            self._log.info('service_url : %s', self._service.addr)

        # always listen on a local port for commands
        self._cmd_port = mt.Thread(target=self.command_port)
        self._cmd_port.daemon = True
        self._cmd_port.start()


    # --------------------------------------------------------------------------
    #
    def _ep_submit_tasks(self, request):

        tasks = request['tasks']

        for task in tasks:

            td  = task['description']
            tid = task.get('uid')

            if not tid:
                tid = ru.generate_id('task.ep.%(item_counter)04d',
                                     ru.ID_CUSTOM, ns=self._pid)

            sbox = self._cfg.pilot_sandbox + '/' + tid

            task['uid']               = tid
            task['origin']            = 'agent'
            task['pilot']             = self._cfg.pid
            task['state']             = rps.AGENT_STAGING_INPUT_PENDING
            task['pilot_sandbox']     = self._cfg.pilot_sandbox
            task['session_sandbox']   = self._cfg.session_sandbox
            task['resource_sandbox']  = self._cfg.resource_sandbox
            task['resources']         = {'cpu': td.ranks * td.cores_per_rank,
                                         'gpu': td.ranks * td.gpus_per_rank}

            task['task_sandbox']      = 'file://localhost/' + sbox
            task['task_sandbox_path'] = sbox

            self._log.debug('ep: submit %s', td['uid'])

        self.advance(tasks, state=rps.AGENT_STAGING_INPUT_PENDING,
                            publish=True, push=True)


  # # --------------------------------------------------------------------------
  # #
  # def _ep_get_task_updates(self, request):
  #
  #     import pprint
  #     self._log.debug('update request: %s', pprint.pformat(request))


# ------------------------------------------------------------------------------





# ------------------------------------------------------------------------------
#
# global constants
#
MASTER                         = 'master'
WORKER                         = 'worker'

STAGER_WORKER                  = 'stager'
STAGER_REQUEST_QUEUE           = 'stager_request_queue'
STAGER_RESPONSE_PUBSUB         = 'stager_response_pubsub'

PMGR_LAUNCHING_QUEUE           = 'pmgr_launching_queue'
PMGR_LAUNCHING_COMPONENT       = 'pmgr_launching'

TMGR_SCHEDULING_QUEUE          = 'tmgr_scheduling_queue'
TMGR_STAGING_INPUT_QUEUE       = 'tmgr_staging_input_queue'
TMGR_STAGING_OUTPUT_QUEUE      = 'tmgr_staging_output_queue'

TMGR_SCHEDULING_COMPONENT      = 'tmgr_scheduling'
TMGR_STAGING_INPUT_COMPONENT   = 'tmgr_staging_input'
TMGR_STAGING_OUTPUT_COMPONENT  = 'tmgr_staging_output'

AGENT_STAGING_INPUT_PUBSUB     = 'agent_staging_input_pubsub'
AGENT_STAGING_INPUT_QUEUE      = 'agent_staging_input_queue'
AGENT_SCHEDULING_QUEUE         = 'agent_scheduling_queue'
AGENT_EXECUTING_QUEUE          = 'agent_executing_queue'
AGENT_STAGING_OUTPUT_QUEUE     = 'agent_staging_output_queue'
AGENT_COLLECTING_QUEUE         = 'agent_collecting_queue'

RAPTOR_SCHEDULING_QUEUE        = 'raptor_scheduling_queue'

AGENT_STAGING_INPUT_COMPONENT  = 'agent_staging_input'
AGENT_SCHEDULING_COMPONENT     = 'agent_scheduling'
AGENT_EXECUTING_COMPONENT      = 'agent_executing'
AGENT_STAGING_OUTPUT_COMPONENT = 'agent_staging_output'

TMGR_UNSCHEDULE_PUBSUB         = 'tmgr_unschedule_pubsub'
TMGR_RESCHEDULE_PUBSUB         = 'tmgr_reschedule_pubsub'

AGENT_UNSCHEDULE_PUBSUB        = 'agent_unschedule_pubsub'
AGENT_SCHEDULE_PUBSUB          = 'agent_schedule_pubsub'

CONTROL_PUBSUB                 = 'control_pubsub'
STATE_PUBSUB                   = 'state_pubsub'
LOG_PUBSUB                     = 'log_pubsub'

PROXY_CONTROL_PUBSUB           = 'proxy_control_pubsub'
PROXY_STATE_PUBSUB             = 'proxy_state_pubsub'
PROXY_TASK_QUEUE               = 'proxy_task_queue'

CLIENT_PUBSUB                  = 'client_pubsub'
CLIENT_QUEUE                   = 'client_queue'


# ------------------------------------------------------------------------------
#
# constants for resource occupation: floats allow for partial occupation
#
FREE =  0.0
BUSY =  1.0
DOWN = None


# -----------------------------------------------------------------------------
#
# definitions of metrics
#
TASK_STATE           = 'TASK_STATE'
WAIT_QUEUE_SIZE      = 'WAIT_QUEUE_SIZE'
TMGR_METRICS         = [TASK_STATE,
                        WAIT_QUEUE_SIZE]

PILOT_STATE          = 'PILOT_STATE'
PMGR_METRICS         = [PILOT_STATE]


# ------------------------------------------------------------------------------
#
# staging defines and flags
#
COPY     = 'Copy'      # local cp
LINK     = 'Link'      # local ln -s
MOVE     = 'Move'      # local mv
DOWNLOAD = 'Download'  # remote download by agent
TRANSFER = 'Transfer'  # remote transfer from / to client
TARBALL  = 'Tarball'   # remote staging will be executed using a tarball.

RECURSIVE      = 1     # recursive copy of directories
CREATE_PARENTS = 2     # Create parent directories if needed
NON_FATAL      = 4     # Don't fail the Task if input is missing

#
# Defaults
#
DEFAULT_ACTION   = TRANSFER
DEFAULT_PRIORITY = 0
DEFAULT_FLAGS    = CREATE_PARENTS


#
# Task MPI flags
#
SERIAL         = 'Serial'
MPI            = 'MPI'
OpenMP         = 'OpenMP'
GPU            = 'GPU'
GPU_MPI        = 'GPU_MPI'
GPU_OpenMP     = 'GPU_OpenMP'

# process / thread types (for both, CPU and GPU processes/threads)
POSIX          = 'POSIX'   # native threads / application threads
CUDA           = 'CUDA'
FUNC           = 'FUNC'


# scheduler names
SCHEDULER_ROUND_ROBIN  = "round_robin"
SCHEDULER_BACKFILLING  = "backfilling"
SCHEDULER_DEFAULT      = SCHEDULER_ROUND_ROBIN


# ------------------------------------------------------------------------------




# pylint: disable=unused-argument

__copyright__ = 'Copyright 2013-2022, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import os
import time
import queue
import atexit
import signal
import threading  as mt
import subprocess as sp

import radical.utils as ru

from ...  import states    as rps
from ...  import constants as rpc

from .base import AgentExecutingComponent


# ------------------------------------------------------------------------------
# ensure tasks are killed on termination
_pids = list()


# pylint: disable=unused-argument
def _kill(*args, **kwargs):

    for pid in _pids:

        # skip test mocks
        if not isinstance(pid, int):
            continue

        try   : os.killpg(pid, signal.SIGTERM)
        except: pass


atexit.register(_kill)
signal.signal(signal.SIGTERM, _kill)
signal.signal(signal.SIGINT,  _kill)
# ------------------------------------------------------------------------------


# ------------------------------------------------------------------------------
#
class Popen(AgentExecutingComponent):

    # --------------------------------------------------------------------------
    #
    def initialize(self):

      # self._log.debug('popen initialize start')
        super().initialize()

        self._tasks       = dict()
        self._check_lock  = mt.Lock()
        self._watch_queue = queue.Queue()

        # run watcher thread
        self._watcher = mt.Thread(target=self._watch)
        self._watcher.daemon = True
        self._watcher.start()


    # --------------------------------------------------------------------------
    #
    def get_task(self, tid):

        return self._tasks.get(tid)


    # --------------------------------------------------------------------------
    #
    def cancel_task(self, task):

        # was the task even started?
        tid  = task['uid']
        proc = task.get('proc')
        if not proc:
            # task was not started, nothing to do
            # FIXME: this is a race, the executor might already own the task
            self._log.debug('task %s was not started', tid)
            return

        # check if the task is, maybe, already done
        exit_code = proc.poll()
        if exit_code is not None:
            # task is done, nothing to do
            self._log.debug('task %s is already done', tid)
            return

        # remove from tasks dictionary, thus "watcher" will not pick it up
        with self._check_lock:
            if tid not in self._tasks:
                return
            try:
                del self._tasks[tid]
            except KeyError:
                pass

        # task is still running -- cancel it
        self._log.debug('cancel %s', tid)
        self._prof.prof('task_run_cancel_start', uid=tid)

        launcher = self._rm.get_launcher(task['launcher_name'])
        launcher.cancel_task(task, proc.pid)

        proc.wait()  # make sure proc is collected

        try:
            # might race with task collection
            del task['proc']  # proc is not json serializable
        except KeyError:
            pass

        task['exit_code']    = None
        task['target_state'] = rps.CANCELED

        self._prof.prof('task_run_cancel_stop', uid=tid)
        self._prof.prof('unschedule_start', uid=tid)
        self.publish(rpc.AGENT_UNSCHEDULE_PUBSUB, task)

        self.advance([task], rps.AGENT_STAGING_OUTPUT_PENDING,
                             publish=True, push=True)


    # --------------------------------------------------------------------------
    #
    def work(self, tasks):

        self.advance_tasks(tasks, rps.AGENT_EXECUTING, publish=True, push=False)

        for task in tasks:

            try:
                self._prof.prof('task_start', uid=task['uid'])
                self._tasks.update({task['uid']: task})
                self._handle_task(task)

            except Exception as e:
                self._log.exception("error running Task")
                task['exception']        = repr(e)
                task['exception_detail'] = '\n'.join(ru.get_exception_trace())

                # can't rely on the executor base to free the task resources
                self._prof.prof('unschedule_start', uid=task['uid'])
                self.publish(rpc.AGENT_UNSCHEDULE_PUBSUB, task)

                self.advance_tasks(task, rps.FAILED, publish=True, push=False)


    # --------------------------------------------------------------------------
    #
    def _handle_task(self, task):

        # before we start handling the task, check if it should run in a named
        # env.  If so, inject the activation of that env in the task's pre_exec
        # directives.
        tid  = task['uid']
        td   = task['description']
        sbox = task['task_sandbox_path']

        # prepare stdout/stderr
        task['stdout'] = ''
        task['stderr'] = ''

        stdout_file    = td.get('stdout') or '%s.out' % tid
        stderr_file    = td.get('stderr') or '%s.err' % tid

        if stdout_file[0] != '/':
            task['stdout_file']       = '%s/%s' % (sbox, stdout_file)
            task['stdout_file_short'] = '$RP_TASK_SANDBOX/%s' % stdout_file
        else:
            task['stdout_file']       = stdout_file
            task['stdout_file_short'] = stdout_file

        if stderr_file[0] != '/':
            task['stderr_file']       = '%s/%s' % (sbox, stderr_file)
            task['stderr_file_short'] = '$RP_TASK_SANDBOX/%s' % stderr_file
        else:
            task['stderr_file']       = stderr_file
            task['stderr_file_short'] = stderr_file

        # create two shell scripts: a launcher script (task.launch.sh) which
        # sets the launcher environment, performs pre_launch commands, and then
        # launches the second script which executes the task.
        #
        # The second script (`task.exec.sh`) is instantiated once per task rank.
        # It first resets the environment created by the launcher, then prepares
        # the environment for the tasks.  Next it runs the `pre_exec` directives
        # for all ranks and per rank (if commands for particular rank are
        # defined) are executed, and then, after all ranks are synchronized,
        # finally the task ranks begin to run.
        #
        # The scripts thus show the following approximate structure:
        #
        # Launcher Script (`task.000000.launch.sh`):
        # ----------------------------------------------------------------------
        # #!/bin/sh
        #
        # # `pre_launch` commands
        # date > data.input
        #
        # # launcher specific environment setup
        # module load openmpi
        #
        # # launch the task script
        # mpirun -n 4 ./task.000000.exec.sh
        # ----------------------------------------------------------------------
        #
        # Task Script (`task.000000.exec.sh`)
        # ----------------------------------------------------------------------
        # #!/bin/sh
        #
        # # clean launch environment
        # module unload mpi
        #
        # # task environment setup (`pre_exec`)
        # module load gromacs
        #
        # # rank specific setup
        # touch task.000000.ranks
        # if test "$MPI_RANK" = 0; then
        #   export CUDA_VISIBLE_DEVICES=0
        #   export OMP_NUM_THREADS=2
        #   export RANK_0_VAR=foo
        #   echo 0 >> task.000000.ranks
        # fi
        #
        # if test "$MPI_RANK" = 1; then
        #   export CUDA_VISIBLE_DEVICES=1
        #   export OMP_NUM_THREADS=4
        #   export RANK_1_VAR=bar
        #   echo 1 >> task.000000.ranks
        # fi
        #
        # # synchronize ranks
        # while $(cat task.000000.ranks | wc -l) != $MPI_RANKS; do
        #   sleep 1
        # done
        #
        # # run the task
        # mdrun -i ... -o ... -foo ... 1> task.000000.$MPI_RANK.out \
        #                              2> task.000000.$MPI_RANK.err
        #
        # # now do the very same stuff for the `post_exec` directive
        # ...

        launcher, lname = self._rm.find_launcher(task)

        if not launcher:
            raise RuntimeError('no launcher found for %s' % task)

        task['launcher_name'] = lname

        exec_path  , _ = self._create_exec_script(launcher, task)
        _, launch_path = self._create_launch_script(launcher, task, exec_path)

        tid  = task['uid']
        sbox = task['task_sandbox_path']

        # launch and exec script are done, get ready for execution.
        self._log.info('Launching task %s via %s in %s', tid, launch_path, sbox)

        _launch_out_h = ru.ru_open('%s/%s.launch.out' % (sbox, tid), 'w')


        # `start_new_session=True` is default, which enables decoupling
        # from the parent process group (part of the task cancellation)
        _start_new_session = self.session.rcfg.new_session_per_task or False

        self._prof.prof('task_run_start', uid=tid)
        task['proc'] = sp.Popen(args              = launch_path,
                                executable        = None,
                                shell             = False,
                                stdin             = None,
                                stdout            = _launch_out_h,
                                stderr            = sp.STDOUT,
                                start_new_session = _start_new_session,
                                close_fds         = True,
                                cwd               = sbox)
        self._prof.prof('task_run_ok', uid=tid)

        # store pid for last-effort termination
        _pids.append(task['proc'].pid)

        # handle task timeout if needed
        self.handle_timeout(task)

        # watch task for completion
        self._watch_queue.put(task)

        # now that the task cancellation cb would succeed, let's make sure that
        # no cancellation request sneaked in before the task got started
        if self.is_canceled(task) is True:
            self.cancel_task(task)


    # --------------------------------------------------------------------------
    #
    def _watch(self):

        to_watch = list()  # contains task dicts

        try:
            while not self._term.is_set():

                # FIXME: we don't want to only wait for one Task -- then we
                #        would pull Task state too frequently.  OTOH, we
                #        also don't want to learn about tasks until all
                #        slots are filled, because then we may not be able
                #        to catch finishing tasks in time -- so there is
                #        a fine balance here.  Fine balance means 100.
                MAX_QUEUE_BULKSIZE = 100
                count = 0

                try:
                    while count < MAX_QUEUE_BULKSIZE:
                        to_watch.append(self._watch_queue.get_nowait())
                        count += 1

                except queue.Empty:
                    pass

                # check on the known tasks.
                self._check_running(to_watch)

                if not count:
                    # no new tasks, no new state -- sleep a bit
                    time.sleep(0.05)

        except Exception as e:
            self._log.exception('Error in ExecWorker watch loop (%s)' % e)
            # FIXME: this should signal the ExecWorker for shutdown...


    # --------------------------------------------------------------------------
    # Iterate over all running tasks, check their status, and decide on the
    # next step.  Also check for a requested cancellation for the tasks.
    def _check_running(self, to_watch):

        tasks_to_advance = list()

        # `to_watch.remove()` in the loop requires copy to iterate over the list
        for task in list(to_watch):

            task_proc = task.get('proc')
            if task_proc is None:
                to_watch.remove(task)
                continue

            tid = task['uid']

            # poll subprocess object
            exit_code = task_proc.poll()
            if exit_code is not None:

                self._prof.prof('task_run_stop', uid=tid)

                # make sure proc is collected
                task_proc.wait()

                # we have a valid return code -- task is final
                self._log.info("Task %s has return code %s.", tid, exit_code)

                # Free the Slots, Flee the Flots, Ree the Frots!
                to_watch.remove(task)

                try:
                    # might race with task cancellation
                    del task['proc']  # proc is not json serializable
                except KeyError:
                    pass

                with self._check_lock:
                    if tid not in self._tasks:
                        # task was canceled before, nothing to do
                        continue
                    try:
                        del self._tasks[tid]
                    except KeyError:
                        pass

                tasks_to_advance.append(task)

                self._prof.prof('unschedule_start', uid=tid)

                if exit_code == 0:
                    # The task finished cleanly, see if we need to deal with
                    # output data.  We always move to stageout, even if there
                    # are no directives -- at the very least, we'll upload
                    # stdout/stderr
                    task['exit_code']    = exit_code
                    task['target_state'] = rps.DONE

                else:
                    # task failed (we still run staging output)
                    task['exit_code']        = exit_code
                    task['exception']        = 'RuntimeError("task failed")'
                    task['exception_detail'] = 'exit code: %s' % exit_code
                    task['target_state']     = rps.FAILED

        self.publish(rpc.AGENT_UNSCHEDULE_PUBSUB, tasks_to_advance)

        if tasks_to_advance:
            self.advance(tasks_to_advance, rps.AGENT_STAGING_OUTPUT_PENDING,
                                           publish=True, push=True)


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2024, http://radical.rutgers.edu"
__license__   = "MIT"

import os
import time

from functools   import partial
from collections import defaultdict

import radical.utils as ru

from .        import constants as rpc
from .session import Session


# ------------------------------------------------------------------------------
#
class Client(object):

    # --------------------------------------------------------------------------
    #
    def __init__(self, uid=None, reg_addr=None):

        self._uid      = uid      or os.environ.get('RP_SESSION_ID')
        self._reg_addr = reg_addr or os.environ.get('RP_REGISTRY_ADDRESS')

        if not self._uid:
            raise ValueError('no session ID given nor found in env')

        if not self._reg_addr:
            raise ValueError('no registry address given nor found in env')

        self._session = Session(uid=self._uid,
                                _role=Session._CLIENT,
                                _reg_addr=self._reg_addr)

        self._log  = self._session._get_logger('radical.client')
        self._prof = self._session._get_profiler('radical.client')

        self._session._reg.dump('client')

        # hook into control and state pubsubs.  On messages, invoke registered
        # callbacks
        self._ctrl_cbs   = defaultdict(list)
        self._state_cbs  = defaultdict(list)
        self._pubsub_cbs = defaultdict(list)
        self._queue_cbs  = defaultdict(list)

        # ----------------------------------------------------------------------
        # control pubsub
        bname    = 'bridges.%s' % rpc.CONTROL_PUBSUB
        ctrl_cfg = ru.Config(self._session._reg[bname])
        self._ctrl_pub = ru.zmq.Publisher(channel=rpc.CONTROL_PUBSUB,
                                          url=ctrl_cfg.addr_pub,
                                          log=self._log,
                                          prof=self._prof)
        self._ctrl_sub = ru.zmq.Subscriber(channel=rpc.CONTROL_PUBSUB,
                                           url=ctrl_cfg.addr_sub,
                                           log=self._log,
                                           prof=self._prof)


        # ----------------------------------------------------------------------
        # state pubsub (no publisher!)
        bname     = 'bridges.%s' % rpc.STATE_PUBSUB
        state_cfg = ru.Config(self._session._reg[bname])
        self._state_sub = ru.zmq.Subscriber(channel=rpc.STATE_PUBSUB,
                                            url=state_cfg.addr_sub,
                                            log=self._log,
                                            prof=self._prof)


        # ----------------------------------------------------------------------
        # client pubsub
        bname      = 'bridges.%s' % rpc.CLIENT_PUBSUB
        pubsub_cfg = ru.Config(self._session._reg[bname])
        self._pub  = ru.zmq.Publisher(channel=rpc.CLIENT_PUBSUB,
                                      url=pubsub_cfg.addr_pub,
                                      log=self._log,
                                      prof=self._prof)
        self._sub = ru.zmq.Subscriber(channel=rpc.CLIENT_PUBSUB,
                                      url=pubsub_cfg.addr_sub,
                                      log=self._log,
                                      prof=self._prof)

        # ----------------------------------------------------------------------
        # client queue
        bname       = 'bridges.%s' % rpc.CLIENT_QUEUE
        queue_cfg   = ru.Config(self._session._reg[bname])
        self._putter = ru.zmq.Putter(channel=rpc.CLIENT_QUEUE,
                                     url=queue_cfg.addr_put,
                                     log=self._log,
                                     prof=self._prof)
        self._getter = ru.zmq.Getter(channel=rpc.CLIENT_QUEUE,
                                     url=queue_cfg.addr_get,
                                     log=self._log,
                                     prof=self._prof)

        # let zmq settle
        time.sleep(1.1)


    # --------------------------------------------------------------------------
    #
    # control pubsub
    #
    def _ctrl_cb(self, topic, msg):

        self._log.debug('control msg %s: %s', topic, msg)
        for cb in self._ctrl_cbs.get(topic, []):
            try:
                ret = cb(topic, msg)
                if ret is False:
                    self._ctrl_cbs[topic].remove(cb)
            except:
                self._log.exception('control cb %s failed' % repr(cb))


    # --------------------------------------------------------------------------
    #
    def register_ctrl_cb(self, topic, cb):

        self._log.debug('register ctrl cb %s: %s', topic, cb)
        self._ctrl_cbs[topic].append(cb)


    # --------------------------------------------------------------------------
    #
    def subscribe_ctrl_topic(self, topic, cb=None):

        if not topic:
            topic = rpc.CONTROL_PUBSUB

        self._log.debug('subscribe to %s', topic)
        self._ctrl_sub.subscribe(topic, self._ctrl_cb)

        if cb:
            self.register_ctrl_cb(topic, cb)


    # --------------------------------------------------------------------------
    #
    def send_ctrl_msg(self, topic, msg):

        self._log.debug('send ctrl msg %s: %s', topic, msg)
        self._ctrl_pub.put(topic, msg)


    # --------------------------------------------------------------------------
    #
    # state pubsub
    #
    def _state_cb(self, topic, msg):

        self._log.debug_5('client state msg %s: %s', topic, msg)
        for cb in self._state_cbs.get(topic, []):
            try:
                ret = cb(topic, msg)
                if ret is False:
                    self._state_cbs[topic].remove(cb)
            except:
                self._log.exception('state cb %s failed' % repr(cb))


    # --------------------------------------------------------------------------
    #
    def subscribe_state_topic(self, topic=None, cb=None):

        if not topic:
            topic = rpc.STATE_PUBSUB

        self._state_sub.subscribe(topic, self._state_cb)

        if cb:
            self.register_state_cb(topic, cb)


    # --------------------------------------------------------------------------
    #
    def register_state_cb(self, topic, cb):

        self._state_cbs[topic].append(cb)


    # --------------------------------------------------------------------------
    #
    # client pubsub
    #
    def _pubsub_cb(self, topic, msg):

        self._log.debug('pubsub msg %s: %s', topic, msg)
        for cb in self._pubsub_cbs.get(topic, []):
            try:
                ret = cb(topic, msg)
                if ret is False:
                    self._pubsub_cbs[topic].remove(cb)
            except:
                self._log.exception('pubsub cb %s failed' % repr(cb))


    # --------------------------------------------------------------------------
    #
    def register_pubsub_cb(self, topic, cb):

        self._pubsub_cbs[topic].append(cb)


    # --------------------------------------------------------------------------
    #
    def subscribe_pubsub_topic(self, topic, cb=None):

        if not topic:
            topic = 'default'

        self._sub.subscribe(topic, self._pubsub_cb)

        if cb:
            self.register_pubsub_cb(topic, cb)


    # --------------------------------------------------------------------------
    #
    def send_pubsub_msg(self, topic, msg):

        self._pub.put(topic, msg)


    # --------------------------------------------------------------------------
    #
    # client queue
    #
    def _queue_cb(self, qname, msg):

        self._log.debug('queue msg %s: %s', qname, msg)
        for cb in self._queue_cbs.get(qname, []):
            try:
                ret = cb(qname, msg)
                if ret is False:
                    self._queue_cbs[qname].remove(cb)
            except:
                self._log.exception('queue cb %s failed' % repr(cb))


    # --------------------------------------------------------------------------
    #
    def register_queue_cb(self, qname, cb):

        # bind class callback to queue name
        self._queue_cbs[qname].append(cb)


    # --------------------------------------------------------------------------
    #
    def subscribe_queue_topic(self, qname, cb=None):

        if not qname:
            qname = 'default'

        bound = partial(self._queue_cb, qname)
        self._getter.subscribe(qname, bound)

        if cb:
            self.register_queue_cb(qname, cb)


    # --------------------------------------------------------------------------
    #
    def send_queue_msg(self, qname, msg):

        self._putter.put(qname, msg)


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

__copyright__ = 'Copyright 2013-2014, http://radical.rutgers.edu'
__license__   = 'MIT'

import sys
import time
import random

import radical.pilot as rp
import radical.utils as ru


# ------------------------------------------------------------------------------
#
def benchmark(reporter):

    reporter.title('pilot resource management - benchmark')

    report.header('benchmark node list creation')

    n_nodes        =   9472
    gpus_per_node  =      8
    cores_per_node =     64
    mem_per_node   =    512
    lfs_per_node   =   1920

    n_tasks        =  10000
    ranks_per_task =      2
    cores_per_rank =      2
    gpus_per_rank  =      1
    mem_per_rank   =      0
    lfs_per_rank   =      0

  # n_nodes        =      2
  # gpus_per_node  =      2
  # cores_per_node =      8
  # mem_per_node   =    512
  # lfs_per_node   =    512
  #
  # n_tasks        =      1
  # ranks_per_task =      2
  # cores_per_rank =      2
  # gpus_per_rank  =      1
  # mem_per_rank   =      0
  # lfs_per_rank   =      0

    start = time.time()
    nodes = [rp.Node({
              'index'   : i,
              'name'    : 'node_%05d' % i,
              'cores'   : [rp.RO(index=x, occupation=rp.FREE)
                                          for x in range(cores_per_node)],
              'gpus'    : [rp.RO(index=x, occupation=rp.FREE)
                                          for x in range(gpus_per_node)],
              'lfs'     : lfs_per_node,
              'mem'     : mem_per_node,
             }) for i in range(n_nodes)]
    nl = rp.NodeList(nodes=nodes)
    stop = time.time()

  # numa_domain_map = {0: rp.NumaDomain(cores=range(0, 4), gpus=[0]),
  #                    1: rp.NumaDomain(cores=range(4, 8), gpus=[1])}
  # nnl = rp.NumaNodeList(nodes=[rp.NumaNode(node, numa_domain_map)
  #                              for node in nodes])
  #
  # import pprint
  # pprint.pprint(nnl.as_dict())
  # pprint.pprint(nnl.nodes[0].as_dict())

    report.ok('nodelist      : %8.2f sec / %d nodes\n' % (stop - start, n_nodes))
    report.ok('                %8.2f nodes / sec\n' % (n_nodes / (stop - start)))

    report.header('benchmark slot allocation')
    rr = rp.RankRequirements(n_cores=cores_per_rank,
                             n_gpus=gpus_per_rank,
                             mem=mem_per_rank,
                             lfs=lfs_per_rank)

    allocs = list()
    start  = time.time()
    hits   = 0
    free   = 0
    miss   = 0
    for i in range(n_tasks):

        slots = nl.find_slots(rr, n_slots=ranks_per_task)
        if slots:
            hits += 1
            allocs.append(slots)
        else:
            miss += 1

        if allocs and random.random() < 0.5:
            free += 1
            to_release = random.choice(allocs)
            allocs.remove(to_release)
            nl.release_slots(to_release)

    stop = time.time()
    report.ok('find slots    : %8.2f sec / %d slots\n'
              % (stop - start, n_tasks * ranks_per_task))
    report.ok('                %8.2f slots / sec\n'
              % (n_tasks * ranks_per_task / (stop - start)))
    report.ok('hits/miss/free: %8d / %d / %d\n' % (hits, miss, free))

    for slots in allocs:
        nl.release_slots(slots)

    report.header('')

  # sys.exit()


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    report = ru.Reporter(name='radical.pilot')

    if True:
        benchmark(report)
        sys.exit()

    report.title('pilot resource management example')

    session = rp.Session()

    try:
        pmgr = rp.PilotManager(session=session)
        tmgr = rp.TaskManager(session=session)

        report.header('submit pilots')

        pd_init = {'resource': 'local.localhost',
                   'runtime' : 15,
                   'nodes'   : 1}
        pdesc = rp.PilotDescription(pd_init)
        pilot = pmgr.submit_pilots(pdesc)

        tmgr.add_pilots(pilot)
        pilot.wait([rp.PMGR_ACTIVE, rp.FAILED])

        n = 1
        report.header('submit %d tasks' % n)

        tds = list()
        for i in range(n):
            slots = pilot.nodelist.find_slots(
                        rp.RankRequirements(n_cores=1, lfs=512), n_slots=2)

            print()
            if slots:
                for slot in slots:
                    print('=== %s' % slot)


            td = rp.TaskDescription()
            td.executable     = '/bin/date'
            td.ranks          = 2
            td.cores_per_rank = 2
            td.slots          = slots

            tds.append(td)

        tasks = tmgr.submit_tasks(tds)
        tmgr.wait_tasks()

        import pprint
        for task in tasks:
            print('  * %s: %s [%s], %s' % (task.uid, task.state, task.exit_code,
                                           task.stdout.strip()))

            print()
            for slot in task.slots:
                print(slot)

    finally:
        report.header('finalize')
        session.close(download=True)

    report.header()


# ------------------------------------------------------------------------------






from typing import Any

import radical.utils as ru


# ------------------------------------------------------------------------------
#
class RPBaseMessage(ru.Message):

    # rpc distinguishes messages which are forwarded to the proxy bridge and
    # those which are not and thus remain local to the module they originate in.

    _msg_type = 'rp_msg'
    _schema   = {'fwd': bool}
    _defaults = {'fwd': False}

    @classmethod
    def register(cls):
        # TODO: this should be moved to the RU base class
        cls._defaults['_msg_type'] = cls._msg_type
        ru.Message.register_msg_type(cls._msg_type, cls)


    # we do not register this message type - it is not supposed to be used
    # directly.


# ------------------------------------------------------------------------------
#
class ComponentStartedMessage(RPBaseMessage):

    # startup messages are never forwarded

    _msg_type = 'component_start'
    _schema   = {'uid': str,
                 'pid': int}
    _defaults = {'fwd': False,
                 'uid': None,
                 'pid': None}


ComponentStartedMessage.register()


# ------------------------------------------------------------------------------
#
class RPCRequestMessage(RPBaseMessage):

    _msg_type = 'rpc_req'
    _schema   = {'uid'      : str,   # uid of message
                 'addr'     : str,   # who is expected to act on the request
                 'cmd'      : str,   # rpc command
                 'args'     : list,  # rpc command arguments
                 'kwargs'   : dict}  # rpc command named arguments
    _defaults = {'fwd'      : True,
                 'uid'      : None,
                 'addr'     : None,
                 'cmd'      : None,
                 'args'     : [],
                 'kwargs'   : {}}


RPCRequestMessage.register()


# ------------------------------------------------------------------------------
#
class RPCResultMessage(RPBaseMessage):

    _msg_type = 'rpc_res'
    _schema   = {'uid'      : str,  # uid of rpc call
                 'val'      : Any,  # return value (`None` by default)
                 'out'      : str,  # stdout
                 'err'      : str,  # stderr
                 'exc'      : str}  # raised exception representation
    _defaults = {'_msg_type': _msg_type,
                 'fwd'      : True,
                 'uid'      : None,
                 'val'      : None,
                 'out'      : None,
                 'err'      : None,
                 'exc'      : None}

    # --------------------------------------------------------------------------
    #
    def __init__(self, rpc_req=None, from_dict=None, **kwargs):

        # when constructed from a request message copy the uid

        if rpc_req:
            if not from_dict:
                from_dict = dict()

            from_dict['uid'] = rpc_req['uid']

        super().__init__(from_dict, **kwargs)


RPCResultMessage.register()


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

# This is an example multithreaded program that is used
# by different examples and tests.

import sys
import time
import socket
import threading


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    host = socket.gethostname().split('.')[0]

    def _thread(size, rank):
        time.sleep(1)
        print("%d/%d/%s"  % (rank + 1, size, host))

    threads = list()
    size    = int(sys.argv[1])
    for n in range(size):
        threads.append(threading.Thread(target=_thread, args=[size, n]))

    for thread in threads:
        time.sleep(0.01)  # ensure ordered output
        thread.start()

    for thread in threads:
        thread.join()

# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2016, http://radical.rutgers.edu"
__license__   = "MIT"


import os

import radical.utils as ru

from .base import LaunchMethod


# ------------------------------------------------------------------------------
#
class SSH(LaunchMethod):

    # --------------------------------------------------------------------------
    #
    def __init__(self, name, lm_cfg, rm_info, log, prof):

        LaunchMethod.__init__(self, name, lm_cfg, rm_info, log, prof)


    # --------------------------------------------------------------------------
    #
    def _init_from_scratch(self, env, env_sh):

        command = ru.which('ssh')

        if not command:
            raise RuntimeError("ssh not found!")

        # Some MPI environments (e.g. SGE) put a link to rsh as "ssh" into
        # the path.  We try to detect that and then use different arguments.
        is_rsh = False
        if os.path.islink(command):

            command = os.path.realpath(command)

            if os.path.basename(command) == 'rsh':
                self._log.info('Detected that "ssh" is a link to "rsh".')
                is_rsh = True

        if not is_rsh:
            command += ' -o StrictHostKeyChecking=no -o ControlMaster=auto'

        lm_info = {'env'    : env,
                   'env_sh' : env_sh,
                   'command': command}

        return lm_info


    # --------------------------------------------------------------------------
    #
    def _init_from_info(self, lm_info):

        self._env     = lm_info['env']
        self._env_sh  = lm_info['env_sh']
        self._command = lm_info['command']

        assert self._command


    # --------------------------------------------------------------------------
    #
    def finalize(self):

        pass


    # --------------------------------------------------------------------------
    #
    def can_launch(self, task):

        # ensure single rank
        if len(task['slots']) > 1:
            return False, 'more than one rank'

        # ensure non-MPI
        if task['description'].get('use_mpi'):
            return False, 'cannot launch MPI tasks'

        if not task['description']['executable']:
            return False, 'no executable'

        return True, ''


    # --------------------------------------------------------------------------
    #
    def get_launcher_env(self):

        return ['. $RP_PILOT_SANDBOX/%s' % self._env_sh]


    # --------------------------------------------------------------------------
    #
    def get_launch_cmds(self, task, exec_path):

        slots = task['slots']

        if len(slots) != 1:
            raise RuntimeError('ssh cannot run multi-rank tasks')

        host = slots[0]['node_name']
        cmd  = '%s %s %s' % (self._command, host, exec_path)
        return cmd.rstrip()


    # --------------------------------------------------------------------------
    #
    def get_rank_cmd(self):

        return 'export RP_RANK=0\n'


# ------------------------------------------------------------------------------





__copyright__ = 'Copyright 2016-2023, The RADICAL-Cybertools Team'
__license__   = 'MIT'

from collections import defaultdict

import radical.utils as ru

from .base import LaunchMethod


# ------------------------------------------------------------------------------
#
class MPIExec(LaunchMethod):

    # --------------------------------------------------------------------------
    #
    def __init__(self, name, lm_cfg, rm_info, log, prof):

        self._mpt    : bool  = False
        self._rsh    : bool  = False
        self._use_rf : bool  = False
        self._use_hf : bool  = False
        self._can_os : bool  = False
        self._ccmrun : str   = ''
        self._dplace : str   = ''
        self._omplace: str   = ''
        self._command: str   = ''

        LaunchMethod.__init__(self, name, lm_cfg, rm_info, log, prof)


    # --------------------------------------------------------------------------
    #
    def _init_from_scratch(self, env, env_sh):

        lm_info = {
            'env'    : env,
            'env_sh' : env_sh,
            'command': ru.which([
                'mpiexec',             # General case
                'mpiexec.mpich',       # Linux, MPICH
                'mpiexec.hydra',       # Linux, MPICH
                'mpiexec.openmpi',     # Linux, MPICH
                'mpiexec-mpich-mp',    # Mac OSX MacPorts
                'mpiexec-openmpi-mp',  # Mac OSX MacPorts
                'mpiexec_mpt',         # Cheyenne (NCAR)
            ]),
            'mpt'    : False,
            'rsh'    : False,
            'use_rf' : False,
            'use_hf' : False,
            'can_os' : False,
            'ccmrun' : '',
            'dplace' : '',
            'omplace': ''
        }

        if not lm_info['command']:
            raise ValueError('mpiexec not found - cannot start MPI tasks')

        if '_mpt' in self.name.lower():
            lm_info['mpt'] = True

        if '_rsh' in self.name.lower():
            lm_info['rsh'] = True

        # do we need ccmrun or dplace?
        if '_ccmrun' in self.name.lower():
            lm_info['ccmrun'] = ru.which('ccmrun')
            assert lm_info['ccmrun']

        if '_dplace' in self.name.lower():
            lm_info['dplace'] = ru.which('dplace')
            assert lm_info['dplace']

        # cheyenne always needs mpt and omplace
        if 'cheyenne' in ru.get_hostname():
            lm_info['omplace'] = 'omplace'
            lm_info['mpt']     = True

        # check that this implementation allows to use `rankfile` option
        lm_info['use_rf'] = self._check_available_lm_options(
            lm_info['command'], '-rf')

        # if we fail, then check if this implementation allows to use
        # `host names` option
        if not lm_info['use_rf']:
            lm_info['use_hf'] = self._check_available_lm_options(
                lm_info['command'], '-f')

        # check that this implementation allows to use `oversubscribe` option
        if self._rm_info.details.get('oversubscribe'):
            lm_info['can_os'] = self._check_available_lm_options(
                lm_info['command'], '--oversubscribe')
            if not lm_info['can_os']:
                self._log.warn('oversubscribe requested but not supported')

        mpi_version, mpi_flavor = self._get_mpi_info(lm_info['command'])
        lm_info['mpi_version']  = mpi_version
        lm_info['mpi_flavor']   = mpi_flavor

        return lm_info

    # --------------------------------------------------------------------------
    #
    def _check_available_lm_options(self, lm_cmd, option):
        cmd   = '%s --help | grep -e "%s\\>"' % (lm_cmd, option)
        check = bool(ru.sh_callout(cmd, shell=True)[0])
        self._log.debug('lm option %s: %s (%s)' % (option, check, cmd))

        if not check:
            # openmpi needs `all` to work here
            cmd   = '%s --help all | grep -e "%s\\>"' % (lm_cmd, option)
            check = bool(ru.sh_callout(cmd, shell=True)[0])
            self._log.debug('lm option %s: %s (%s)' % (option, check, cmd))

        if not check:
            # openmpi is different nowadays...
            check = bool(ru.sh_callout('%s --help mapping | grep -e "%s\\>"' %
                                      (lm_cmd, option), shell=True)[0])

        return check

    # --------------------------------------------------------------------------
    #
    def _init_from_info(self, lm_info):

        self._env         = lm_info['env']
        self._env_sh      = lm_info['env_sh']
        self._command     = lm_info['command']

        assert self._command

        self._mpt         = lm_info['mpt']
        self._rsh         = lm_info['rsh']
        self._use_rf      = lm_info['use_rf']
        self._use_hf      = lm_info['use_hf']
        self._can_os      = lm_info['can_os']
        self._dplace      = lm_info['dplace']
        self._ccmrun      = lm_info['ccmrun']

        self._mpi_version = lm_info['mpi_version']
        self._mpi_flavor  = lm_info['mpi_flavor']

        # ensure empty string on unset omplace
        if not lm_info['omplace']:
            self._omplace = ''
        else:
            self._omplace = 'omplace'


    # --------------------------------------------------------------------------
    #
    def finalize(self):

        pass


    # --------------------------------------------------------------------------
    #
    def can_launch(self, task):

        if not task['description']['executable']:
            return False, 'no executable'

        return True, ''


    # --------------------------------------------------------------------------
    #
    def get_launcher_env(self):

        lm_env_cmds = ['. $RP_PILOT_SANDBOX/%s' % self._env_sh]

        # Cheyenne is the only machine that requires mpiexec_mpt.
        # We then have to set MPI_SHEPHERD=true
        if self._mpt:
            lm_env_cmds.append('export MPI_SHEPHERD=true')

        return lm_env_cmds


    # --------------------------------------------------------------------------
    #
    @staticmethod
    def _get_rank_file(slots, uid, sandbox):
        '''
        Rank file:
            rank 0=localhost slots=0,1,2,3
            rank 1=localhost slots=4,5,6,7
        '''
        rf_str  = ''
        rank_id = 0

        for slot in slots:
            rf_str += 'rank %d=%s ' % (rank_id, slot['node_name'])
            rf_str += 'slots=%s\n' % ','.join([str(c['index'])
                                               for c in slot['cores']])
            rank_id += 1

        rf_name = '%s/%s.rf' % (sandbox, uid)
        with ru.ru_open(rf_name, 'w') as fout:
            fout.write(rf_str)

        return rf_name


    # --------------------------------------------------------------------------
    #
    @staticmethod
    def _get_host_file(slots, uid, sandbox, mode=0):
        '''
        Create a hostfile for the given slots
        mode == 0:
            localhost
        mode == 1:
            localhost slots=2
        mode == 2:
            localhost:2
        '''

        host_slots = defaultdict(int)

        for slot in slots:
            host_slots[slot['node_name']] += 1

        if mode == 0:
            hf_str = '%s\n' % '\n'.join(list(host_slots.keys()))

        else:
            hf_str = ''
            if mode == 1: slots_ref = ' slots='
            else        : slots_ref = ':'

            for host_name, num_slots in host_slots.items():
                hf_str += '%s%s%d\n' % (host_name, slots_ref, num_slots)

        hf_name = '%s/%s.hf' % (sandbox, uid)
        with ru.ru_open(hf_name, 'w') as fout:
            fout.write(hf_str)

        return hf_name


    # --------------------------------------------------------------------------
    #
    def get_launch_cmds(self, task, exec_path):

        uid   = task['uid']
        slots = task['slots']
        sbox  = task['task_sandbox_path']

        assert slots, 'task.slots not defined'

        host_slots = defaultdict(int)
        for slot in slots:
            host_slots[slot['node_name']] += 1

        cmd_options = '-np %d ' % sum(host_slots.values())

        if self._use_rf:
            rankfile     = self._get_rank_file(slots, uid, sbox)
            hosts        = set([slot['node_name'] for slot in slots])
          # cmd_options += '-H %s -rf %s ' % (','.join(hosts), rankfile)
            cmd_options += '-rf %s ' % rankfile

        elif self._mpi_flavor == self.MPI_FLAVOR_PALS:
            hostfile     = self._get_host_file(slots, uid, sbox)

            tmp = list()
            for slot in slots:
                cores = slot['cores']
                if len(cores) > 1:
                    tmp.append('%s-%s' % (cores[0]['index'], cores[-1]['index']))
                else:
                    tmp.append(str(cores[0]['index']))
            core_ids = ':'.join(tmp)

          # # FIXME: make this readable please
          # core_ids     = ':'.join([
          #     str(cores[0]) + ('-%s' % cores[-1] if len(cores) > 1 else '')
          #     for core_map in [slot['cores'] for slot in slots]
          #     for cores    in core_map])
            cmd_options += '--ppn %d '           % max(host_slots.values()) + \
                           '--cpu-bind list:%s ' % core_ids + \
                           '--hostfile %s '      % hostfile

            # NOTE: Option "--ppn" controls "node-depth" vs. "core-depth"
            #       process placement. If we submit "mpiexec" command with
            #       "--ppn" option, it will place processes within the same
            #       node first. If we do not provide "--ppn" option, it will
            #       place processes on the available nodes one by one and
            #       round-robin when each available node is populated.

            # if over-subscription is allowed,
            # then the following approach is applicable too:
            #    cores_per_rank = len(slots[0]['cores'])
            #    cmd_options   += '--depth=%d --cpu-bind depth ' % cores_per_rank

        elif self._use_hf:
            hostfile = self._get_host_file(slots, uid, sbox, mode=2)
            cmd_options += '-f %s ' % hostfile
        else:
            hostfile     = self._get_host_file(slots, uid, sbox, mode=1)
            cmd_options += '--hostfile %s ' % hostfile


        if self._rm_info.details.get('oversubscribe') and self._can_os:
            cmd_options += '--oversubscribe '

        if self._omplace:
            cmd_options += '%s ' % self._omplace

        cmd = '%s %s %s' % (self._command, cmd_options, exec_path)
        return cmd.strip()


    # --------------------------------------------------------------------------
    #
    def get_rank_cmd(self):

        # FIXME: we know the MPI flavor, so make this less guesswork

        ret  = 'test -z "$MPI_RANK"     || export RP_RANK=$MPI_RANK\n'
        ret += 'test -z "$PMIX_RANK"    || export RP_RANK=$PMIX_RANK\n'

        if self._mpt:
            ret += 'test -z "$MPT_MPI_RANK" || export RP_RANK=$MPT_MPI_RANK\n'

        if self._mpi_flavor == self.MPI_FLAVOR_HYDRA:
            ret += 'test -z "$PMI_ID"       || export RP_RANK=$PMI_ID\n'
            ret += 'test -z "$PMI_RANK"     || export RP_RANK=$PMI_RANK\n'
        elif self._mpi_flavor == self.MPI_FLAVOR_PALS:
            ret += 'test -z "$PALS_RANKID"  || export RP_RANK=$PALS_RANKID\n'

        return ret


    # --------------------------------------------------------------------------
    #
    def get_env_blacklist(self):
        '''
        extend the base blacklist by SLURM specific environment variables
        '''

        ret = super().get_env_blacklist()

        ret.extend([
                    'OMPI_*',
                    'PMIX_*',
                   ])

        return ret


# ------------------------------------------------------------------------------





import os
import shutil
import requests

import radical.utils as ru

from ..constants import COPY, LINK, MOVE, TRANSFER, DOWNLOAD
from ..constants import TARBALL  # , CREATE_PARENTS, RECURSIVE


# ------------------------------------------------------------------------------
#
class StagingHelper(object):

    def __init__(self, log):

        self._log  = log

        try   : self._backend = StagingHelper_SAGA (self._log)
        except: self._backend = StagingHelper_Local(self._log)

        log.debug('using staging backend %s' % self._backend.__class__.__name__)


    def mkdir(self, tgt, flags=None):
        self._log.debug('mkdir %s', tgt)
        self._backend.mkdir(tgt, flags)

    def rmdir(self, tgt, flags=None):
        self._log.debug('rmdir %s', tgt)
        self._backend.rmdir(tgt, flags)

    def copy(self, src, tgt, flags=None):
        self._log.debug('copy  %s %s', src, tgt)
        self._backend.copy(src, tgt, flags)

    def move(self, src, tgt, flags=None):
        self._log.debug('move  %s %s', src, tgt)
        self._backend.move(src, tgt, flags)

    def link(self, src, tgt, flags=None):
        self._log.debug('link  %s %s', src, tgt)
        self._backend.link(src, tgt, flags)

    def download(self, src, tgt, flags=None):
        self._log.debug('download %s %s', src, tgt)
        self._backend.download(src, tgt, flags)

    def delete(self, tgt, flags=None):
        self._log.debug('rm    %s', tgt)
        self._backend.delete(tgt, flags)

    def sh_callout(self, url, cmd):
        self._log.debug('shcmd %s %s', url, cmd)
        return self._backend.sh_callout(url, cmd)

    def handle_staging_directive(self, sd):

        action  = sd['action']
        src     = sd['source']
        tgt     = sd['target']
        flags   = sd.get('flags', 0)

        assert action in [COPY, LINK, MOVE, TRANSFER, DOWNLOAD]

        self._log.info('%-10s %s', action, src)
        self._log.info('%-10s %s', '', tgt)

        if action in [COPY, TRANSFER]:
            self.copy(src, tgt, flags)

        elif action == LINK:
            self.link(src, tgt, flags)

        elif action == MOVE:
            self.move(src, tgt, flags)

        elif action in [DOWNLOAD]:
            self.download(src, tgt, flags)


# ------------------------------------------------------------------------------
#
class StagingHelper_Local(object):

    def __init__(self, log):
        self._log = log

    def mkdir(self, tgt, flags):
        self._log.debug('mkdir %s', tgt)
        tgt = ru.Url(tgt).path
        ru.rec_makedir(tgt)

    def rmdir(self, tgt, flags):
        tgt = ru.Url(tgt).path
        os.rmdir(tgt)

    def copy(self, src, tgt, flags):
        src = ru.Url(src).path
        tgt = ru.Url(tgt).path
        self.mkdir(os.path.dirname(tgt), flags)
        ru.sh_callout('cp -r %s %s' % (src, tgt))

    def move(self, src, tgt, flags):
        src = ru.Url(src).path
        tgt = ru.Url(tgt).path
        self.mkdir(os.path.dirname(tgt), flags)
        shutil.move(src, tgt)

    def link(self, src, tgt, flags):
        src = ru.Url(src).path
        tgt = ru.Url(tgt).path
        self.mkdir(os.path.dirname(tgt), flags)
        os.link(src, tgt)

    def download(self, src, tgt, flags):
        tgt = ru.Url(tgt).path
        self.mkdir(os.path.dirname(tgt), flags)
        r = requests.get(src, stream=True)
        with open(tgt, 'wb') as fout:
            for chunk in r.iter_content():
                fout.write(chunk)

    def delete(self, tgt, flags):
        tgt = ru.Url(tgt).path
        try   : os.unlink(tgt)
        except: pass

    def sh_callout(self, url, cmd):
        return ru.sh_callout(cmd, shell=True)


# ------------------------------------------------------------------------------
#
class StagingHelper_SAGA(object):

    try:
        import radical.saga.filesystem      as _rsfs
        import radical.saga.utils.misc      as _rsum
        import radical.saga.utils.pty_shell as _rsup
        _has_saga = True
    except:
        _has_saga = False

    def __init__(self, log):
        self._log = log
        if not self._has_saga:
            raise Exception('SAGA-Python not available')

    def mkdir(self, tgt, flags):
        assert self._has_saga


    def rmdir(self, tgt, flags):
        assert self._has_saga


    def copy(self, src, tgt, flags):

        src = ru.Url(src)
        tgt = ru.Url(tgt)

      # # FIXME: why??
      # flags = 0

        src = ru.Url(src)
        tgt = ru.Url(tgt)

        assert self._has_saga

        tmp      = ru.Url(tgt)
        tmp.path = '/'

        fs     = self._rsfs.Directory(str(tmp))
        flags |= self._rsfs.CREATE_PARENTS

        if os.path.isdir(src.path) or src.path.endswith('/'):
            flags |= self._rsfs.RECURSIVE

      # self._log.debug("copy %s 1 -> %s [%s]" % (src, tgt, flags))
        fs.copy(src, tgt, flags=flags)

    def move(self, src, tgt, flags):
        assert self._has_saga


    def link(self, src, tgt, flags):
        assert self._has_saga

    def download(self, src, tgt, flags):
        assert self._has_saga

        self.copy(src, tgt, flags)


    def delete(self, tgt, flags):
        assert self._has_saga

    def sh_callout(self, url, cmd):
        assert self._has_saga

        js_url = ru.Url(url)
        elems  = js_url.schema.split('+')

        if   'ssh'    in elems: js_url.schema = 'ssh'
        elif 'gsissh' in elems: js_url.schema = 'gsissh'
        elif 'fork'   in elems: js_url.schema = 'fork'
        elif len(elems) == 1  : js_url.schema = 'fork'
        else: raise Exception("invalid schema: %s" % js_url.schema)

        if js_url.schema == 'fork':
            js_url.host = 'localhost'

        self._log.debug("_rsup.PTYShell('%s')", js_url)
        shell = self._rsup.PTYShell(js_url)

        ret, out, err = shell.run_sync(cmd)

        return out, err, ret


# ------------------------------------------------------------------------------

